/***************************************************************************

    M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse

    This file is part of MAME32, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

/***************************************************************************

  options.c

  Stores global options and per-game options;

****************************************************************************/

#include "driver.h"

#include <windowsx.h>
#include <windows.h>
#include <winreg.h>
#include <commctrl.h>
#include <assert.h>
#include <direct.h>
#include "mame32.h"
#include "m32util.h"
#include <resource.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <sys/stat.h>
#include "DDrawDisplay.h"
#include "resource.h"
#include "win32text.h"
#include "drivers/neogeo.h"

/***************************************************************************
    Internal function prototypes
 ***************************************************************************/

static void  LoadOptions(void);
static void  SaveOptions(void);
static void  SaveBackupOptions(void);

static void  SaveCfgSettingsOptions(void);
static void  LoadCfgSettingsOptions(char *szVersion);

static void  LoadCfgGameOptions(options_type *o);
static void  SaveCfgGameOptions(options_type *o);

static void  SaveFolderFlags(char *folderName, DWORD dwFlags);

static void  ColumnEncodeString(void *data, char *str);
static void  ColumnDecodeString(const char *str, void *data);

static void  ColumnDecodeWidths(const char *ptr, void *data);

static void  SplitterEncodeString(void *data, char *str);
static void  SplitterDecodeString(const char *str, void *data);

static void  ListEncodeString(void *data, char *str);
static void  ListDecodeString(const char *str, void *data);

static void  FontEncodeString(void *data, char *str);
static void  FontDecodeString(const char *str, void *data);

static BOOL CfgCreate(void);
static void CfgDestroy(void);
static void CfgReset(void);
static BOOL CfgReadDefaultFile(FILE *fp, char *linebuf);
static void CfgReadFile(FILE *fp, char *linebuf);
static void CfgWriteFile(FILE *fp, const char *keyname);
static long CfgQueryValue(const char *name, void *value, DWORD *size);
static void CfgSetValue(const char *name, void *value, DWORD size);
static void CfgSetValue_ex(const char *name);

static DWORD GetCfgOption(char *name);
static void PutCfgOption(char *name, DWORD value);

static BOOL GetCfgBoolOption(char *name);
static void PutCfgBoolOption(char *name, BOOL value);

static char *GetCfgStringOption(char *name);
static void PutCfgStringOption(char *name, char *option);

static void GetCfgObj(CFG_OPTIONS *cfgOpt);
static void PutCfgObj(CFG_OPTIONS *cfgOpt);

INLINE int GetIndexByName(const char *name);

extern void neogeo_set_driver(int game);
extern BOOL GameUsesRasterDriver(int game);

/***************************************************************************
    Internal defines
 ***************************************************************************/

/* Used to create/retrieve cfg database */
#ifdef BUILD_ENGLISH
#define DEFAULT_CFG   "nmamex2e.cfg"
#else
#define DEFAULT_CFG   "nmamex2.cfg"
#endif
#define GUIBACK_CFG   "guiback.cfg"
#define GAMEBACK_CFG  "gameback.cfg"
#define FOLDERS_CFG   "folders.cfg"
#define GAMES_CFG     "games.cfg"

#define LINEBUF_SIZE  65536

/***************************************************************************
    Internal structures
 ***************************************************************************/

typedef struct cfg_t {
    char *name;
    char *value;
} cfg_struct;

static cfg_struct *cfg;

/***************************************************************************
    Internal variables
 ***************************************************************************/
static settings_type settings;

static options_type gOpts;  /* Used when saving/loading from Registry */
static options_type global; /* Global 'default' options */
static options_type *game;  /* Array of Game specific options */

/* Global UI options */
CFG_OPTIONS cfgSettings[] = {
    {"DefaultGame",     CO_STRING,  settings.default_game,      0, 0},
    {"FolderID",        CO_INT,     &settings.folder_id,        0, 0},
    {"ShowScreenShot",  CO_BOOL,    &settings.show_screenshot,  0, 0},
    {"ShowFlyer",       CO_INT,     &settings.show_pict_type,   0, 0},
    {"ShowToolBar",     CO_BOOL,    &settings.show_toolbar,     0, 0},
    {"ShowStatusBar",   CO_BOOL,    &settings.show_statusbar,   0, 0},
    {"ShowFolderList",  CO_BOOL,    &settings.show_folderlist,  0, 0},
    {"GameCheck",       CO_BOOL,    &settings.game_check,       0, 0},
    {"VersionCheck",    CO_BOOL,    &settings.version_check,    0, 0},
    {"MMXCheck",        CO_BOOL,    &settings.mmx_check,        0, 0},
#ifndef BUILD_ENGLISH
    {"UseJapaneseList", CO_BOOL,    &settings.use_japanese_list,0, 0},
#endif
    {"UseModifyThe",    CO_BOOL,    &settings.modify_the,       0, 0},
#ifdef KAILLERA
	{"UseIMEinChat",    CO_BOOL,    &settings.use_ime_in_chat,  0, 0},
	{"UseFixedForm",    CO_BOOL,    &settings.use_fixed_form,   0, 0},
#endif
    {"VersionName",     CO_STRING,  &settings.version_name,     0, 0},
    {"SortColumn",      CO_INT,     &settings.sort_column,      0, 0},
    {"SortReverse",     CO_BOOL,    &settings.sort_reverse,     0, 0},
    {"GUIPosX",         CO_INT,     &settings.area.x,           0, 0},
    {"GUIPosY",         CO_INT,     &settings.area.y,           0, 0},
    {"GUIWidth",        CO_INT,     &settings.area.width,       0, 0},
    {"GUIHeight",       CO_INT,     &settings.area.height,      0, 0},

    {"RomDirs",         CO_PSTRING, &settings.romdirs,          0, 0},
    {"CfgDir",          CO_PSTRING, &settings.cfgdir,           0, 0},
    {"InpDir",          CO_PSTRING, &settings.inpdir,           0, 0},
    {"ImgDir",          CO_PSTRING, &settings.snapdir,          0, 0},
    {"IconDir",         CO_PSTRING, &settings.icondir,          0, 0},
    {"StateDir",        CO_PSTRING, &settings.statedir,         0, 0},
    {"MemcardDir",      CO_PSTRING, &settings.memcarddir,       0, 0},
    {"FlyerDir",        CO_PSTRING, &settings.flyerdir,         0, 0},
    {"CabinetDir",      CO_PSTRING, &settings.cabinetdir,       0, 0},
    {"MarqueeDir",      CO_PSTRING, &settings.marqueedir,       0, 0},
    {"NVRAMDir",        CO_PSTRING, &settings.nvramdir,         0, 0},
    {"ComboDir",        CO_PSTRING, &settings.combodir,         0, 0},
    {"WavDir",          CO_PSTRING, &settings.wavdir,           0, 0},
    /* ListMode needs to be before ColumnWidths settings */
    {"ListMode",        CO_ENCODE,  &settings.view,             ListEncodeString,       ListDecodeString},
    {"Splitters",       CO_ENCODE,  settings.splitter,          SplitterEncodeString,   SplitterDecodeString},
    {"ListFont",        CO_ENCODE,  &settings.list_font,        FontEncodeString,       FontDecodeString},
    {"ColumnWidths",    CO_ENCODE,  &settings.column_width,     ColumnEncodeString,     ColumnDecodeWidths},
    {"ColumnOrder",     CO_ENCODE,  &settings.column_order,     ColumnEncodeString,     ColumnDecodeString},
    {"ColumnShown",     CO_ENCODE,  &settings.column_shown,     ColumnEncodeString,     ColumnDecodeString},
};


/* Game Options */
CFG_OPTIONS cfgGameOpts[] = {
    /* boolean */
    {"IsWindow",            CO_BOOL,    &gOpts.is_window,         0, 0},
    {"BestFit",             CO_BOOL,    &gOpts.display_best_fit,  0, 0},
    {"TripleBuffer",        CO_BOOL,    &gOpts.triple_buffer,     0, 0},
    {"VsyncFullScreen",     CO_BOOL,    &gOpts.vsync_fullscreen,  0, 0},
    {"VsyncWindow",         CO_BOOL,    &gOpts.vsync_window,      0, 0},
    {"UseSystemMemory",     CO_BOOL,    &gOpts.systemmemory,      0, 0},

    {"Stereo",              CO_BOOL,    &gOpts.stereo,            0, 0},
    {"DirectSoundFix",      CO_BOOL,    &gOpts.directsound_fix,   0, 0},

    {"UseJoystick",         CO_BOOL,    &gOpts.use_joystick,      0, 0},
    {"UseAnalogInputMouse", CO_BOOL,    &gOpts.use_ai_mouse,      0, 0},
    {"DiKeyboard",          CO_BOOL,    &gOpts.di_keyboard,       0, 0},
    {"DiJoystick",          CO_BOOL,    &gOpts.di_joystick,       0, 0},

    {"Cheat",               CO_BOOL,    &gOpts.cheat,             0, 0},
    {"AutoPause",           CO_BOOL,    &gOpts.auto_pause,        0, 0},
    {"DisableMMX",          CO_BOOL,    &gOpts.disable_mmx,       0, 0},
    {"ErrorLog",            CO_BOOL,    &gOpts.error_log,         0, 0},
    {"DoNotSleep",          CO_BOOL,    &gOpts.dontsleep,         0, 0},
    {"DisablePatchROMs",    CO_BOOL,    &gOpts.patch_roms,        0, 0},
    {"CycleSkipCpu",        CO_BOOL,    &gOpts.cyclecpu,          0, 0},
	{"DisableNGLO",         CO_BOOL,    &gOpts.nglo_disabled,     0, 0},

    /* integer */
    {"Monitor",             CO_INT,     &gOpts.display_monitor,   0, 0},
    {"FrameSkip",           CO_INT,     &gOpts.frame_skip,        0, 0},
    {"RenderFullScreen",    CO_INT,     &gOpts.render_fullscreen, 0, 0},
    {"RenderWin",           CO_INT,     &gOpts.render_window,     0, 0},
    {"Width",               CO_INT,     &gOpts.width,             0, 0},
    {"Height",              CO_INT,     &gOpts.height,            0, 0},
    {"Brightness",          CO_INT,     &gOpts.brightness,        0, 0},

    {"Sound",               CO_INT,     &gOpts.sound,             0, 0},
    {"SampleRate",          CO_INT,     &gOpts.sample_rate,       0, 0},
    {"Volume",              CO_INT,     &gOpts.volume,            0, 0},

    {"DefaultInput",        CO_INT,     &gOpts.default_input,     0, 0},
    {"JoyID1",              CO_INT,     &gOpts.dijoyid[0],        0, 0},
    {"JoyID2",              CO_INT,     &gOpts.dijoyid[1],        0, 0},

    {"NeoGeoCoreType",      CO_INT,     &gOpts.core_type,         0, 0},
    {"NeoGeoDriverType",    CO_INT,     &gOpts.driver_type,       0, 0},
    {"NeoGeoBIOSType",      CO_INT,     &gOpts.bios_type,         0, 0},

	{"Transparency",        CO_INT,     &gOpts.transparency,      0, 0},
    {"DSoundBufferSize",    CO_INT,     &gOpts.dsound_buffer,     0, 0},
	{"RandomTime",          CO_INT,     &gOpts.randomtime,        0, 0},

    /* double */
    {"Gamma",               CO_DOUBLE,  &gOpts.gamma,             0, 0},
};

#define NUM_SETTINGS (sizeof(cfgSettings) / sizeof(cfgSettings[0]))
#define NUM_GAMEOPTS (sizeof(cfgGameOpts) / sizeof(cfgGameOpts[0]))
#define NUM_EXTRA        (64)
#define CFG_MAX_ENTRIES  (NUM_SETTINGS + NUM_GAMEOPTS + NUM_EXTRA)

static int  num_games = 0;
static BOOL bResetGUI = FALSE;
static BOOL bResetGameDefs = FALSE;

// Default sizes based on 8pt font w/sort arrow in that column
static int default_column_width[] = { 256, 84, 88, 64, 64, 64, 64,256 };
static int default_column_shown[] = {   1,  1,  1,  1,  1,  1,  1,  1 };
// Hidden columns need to go at the end of the order array
static int default_column_order[] = {   0,  1,  4,  5,  3,  7,  6,  2 };

static char *view_modes[VIEW_MAX] = { "Large Icons", "Small Icons", "List", "Details" };

#define DEFAULT_GAME "nam1975"

/***************************************************************************
    External functions
 ***************************************************************************/

void OptionsInit(int total_games)
{
    int i;

    num_games = total_games;

    strcpy(settings.default_game, DEFAULT_GAME);
    settings.folder_id       = 0;
    settings.view            = VIEW_REPORT;
    settings.show_folderlist = TRUE;
    settings.show_toolbar    = TRUE;
    settings.show_statusbar  = TRUE;
    settings.show_screenshot = TRUE;
    settings.game_check      = TRUE;
    settings.version_check   = TRUE;
    settings.mmx_check       = TRUE;
#ifdef BUILD_ENGLISH
    settings.use_japanese_list = FALSE;
#else
    settings.use_japanese_list = TRUE;
#endif
#ifdef KAILLERA
    settings.use_ime_in_chat = FALSE;
    settings.use_fixed_form  = FALSE;
#endif
    memset(settings.version_name, 0, VERSIONTEXT_LEN * sizeof(char));

    for (i = 0; i < COLUMN_MAX; i++)
    {
        settings.column_width[i] = default_column_width[i];
        settings.column_order[i] = default_column_order[i];
        settings.column_shown[i] = default_column_shown[i];
    }

    settings.sort_column = 0;
    settings.sort_reverse= FALSE;
    settings.area.x      = 0;
    settings.area.y      = 0;
    settings.area.width  = 640;
    settings.area.height = 400;
    settings.splitter[0] = 150;
    settings.splitter[1] = 300;

    settings.romdirs     = strdup(".;roms");
    settings.cfgdir      = strdup("nmx2cfg");
    settings.inpdir      = strdup("inp");
    settings.snapdir     = strdup("snap");
    settings.icondir     = strdup("icons");
    settings.statedir    = strdup("sta");
    settings.memcarddir  = strdup("memcard");
    settings.flyerdir    = strdup("flyers");
    settings.cabinetdir  = strdup("cabinets");
    settings.marqueedir  = strdup("marquees");
    settings.nvramdir    = strdup("nvram");
    settings.combodir    = strdup("combo");
    settings.wavdir      = strdup("wav");

#ifdef BUILD_ENGLISH
    settings.list_font.lfHeight         = -8;
#else
    settings.list_font.lfHeight         = -12;
#endif
    settings.list_font.lfWidth          = 0;
    settings.list_font.lfEscapement     = 0;
    settings.list_font.lfOrientation    = 0;
    settings.list_font.lfWeight         = FW_NORMAL;
    settings.list_font.lfItalic         = FALSE;
    settings.list_font.lfUnderline      = FALSE;
    settings.list_font.lfStrikeOut      = FALSE;
#ifdef BUILD_ENGLISH
    settings.list_font.lfCharSet        = ANSI_CHARSET;
#else
    settings.list_font.lfCharSet        = SHIFTJIS_CHARSET;
#endif
    settings.list_font.lfOutPrecision   = OUT_DEFAULT_PRECIS;
    settings.list_font.lfClipPrecision  = CLIP_DEFAULT_PRECIS;
    settings.list_font.lfQuality        = DEFAULT_QUALITY;
    settings.list_font.lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;

#ifdef BUILD_ENGLISH
    strcpy(settings.list_font.lfFaceName, "MS Sans Serif");
#else
    strcpy(settings.list_font.lfFaceName, "lr oSVbN");
#endif

    settings.list_font_color = (COLORREF)-1;

    global.use_default = FALSE;

    global.is_window         = FALSE;
    global.display_best_fit  = TRUE;
    global.display_monitor   = 0;    /* Primary Display */
    global.width             = 640;
    global.height            = 480;
    global.triple_buffer     = FALSE;
    global.vsync_fullscreen  = FALSE;
    global.vsync_window      = FALSE;
    global.systemmemory      = FALSE;
    global.frame_skip        = 0;   /* auto frame skip */
    global.render_fullscreen = 6;   /* HScanLines (no blit) */
    global.render_window     = 6;   /* HScanLines (no blit) */
    global.gamma             = 1.0;
    global.brightness        = 100;

    global.use_joystick  = FALSE;
    global.use_ai_mouse  = TRUE;
    global.di_keyboard   = TRUE;
    global.di_joystick   = TRUE;
    global.default_input = INPUT_LAYOUT_STD;
    global.dijoyid[0]    = 0;
    global.dijoyid[1]    = 1;

#ifndef NOMIDAS
    if (OnNT() == TRUE)
        global.sound = SOUND_MIDAS;
    else
#endif
        global.sound = SOUND_DIRECT;
    global.sample_rate     = 22050;
    global.volume          = 0;
    global.stereo          = TRUE;
    global.directsound_fix = TRUE;

    global.cheat         = FALSE;
    global.auto_pause    = TRUE;
    global.disable_mmx   = TRUE;
    global.dontsleep     = TRUE;
    global.error_log     = FALSE;
	global.cyclecpu      = TRUE;

    global.core_type     = 0;
    global.driver_type   = 0;
    global.bios_type     = BIOS_TYPE_STANDARD;
    global.patch_roms    = FALSE;
	global.nglo_disabled = FALSE;
	global.transparency  = 50;
    global.dsound_buffer = 100;
	global.randomtime    = 60;

    global.play_count   = 0;
    global.has_roms     = UNKNOWN;
    global.is_favorite  = FALSE;

    /* This allocation should be checked */
    game = (options_type *)MyMalloc(num_games * sizeof(options_type));

    for (i = 0; i < num_games; i++)
    {
        game[i] = global;
        game[i].use_default = TRUE;
    }

    CfgCreate();
    SaveBackupOptions();
    LoadOptions();
}

void OptionsExit(void)
{
    SaveOptions();
    MyFree(game);
    MyFree(settings.romdirs);
    MyFree(settings.cfgdir);
    MyFree(settings.inpdir);
    MyFree(settings.snapdir);
    MyFree(settings.icondir);
    MyFree(settings.statedir);
    MyFree(settings.memcarddir);
    MyFree(settings.flyerdir);
    MyFree(settings.cabinetdir);
    MyFree(settings.marqueedir);
    MyFree(settings.nvramdir);
    MyFree(settings.combodir);
    MyFree(settings.wavdir);
    CfgDestroy();
}

options_type * GetDefaultOptions(void)
{
    return &global;
}

options_type * GetGameOptions(int num_game)
{
    int play_count;
    int has_roms;
    BOOL is_favorite;

    assert(0 <= num_game && num_game < num_games);

    neogeo_set_driver(num_game);

    play_count   = game[num_game].play_count;
    has_roms     = game[num_game].has_roms;
    is_favorite  = game[num_game].is_favorite;

    if (game[num_game].use_default)
    {
        game[num_game]              = global;
        game[num_game].use_default  = TRUE;
        game[num_game].play_count   = play_count;
        game[num_game].has_roms     = has_roms;
        game[num_game].is_favorite  = is_favorite;
        game[num_game].driver_type  = GameUsesRasterDriver(num_game) ? 1 : 0;
    }

    return &game[num_game];
}

void ResetGUI(void)
{
    bResetGUI = TRUE;
}

void SetViewMode(int val)
{
    settings.view = val;
}

int GetViewMode(void)
{
    return settings.view;
}

void SetGameCheck(BOOL game_check)
{
    settings.game_check = game_check;
}

BOOL GetGameCheck(void)
{
    return settings.game_check;
}

void SetVersionCheck(BOOL version_check)
{
    settings.version_check = version_check;
}

BOOL GetVersionCheck(void)
{
    return settings.version_check;
}

void SetMMXCheck(BOOL mmx_check)
{
    settings.mmx_check = mmx_check;
}

BOOL GetMMXCheck(void)
{
    return settings.mmx_check;
}

void SetSavedFolderID(UINT val)
{
    settings.folder_id = val;
}

UINT GetSavedFolderID(void)
{
    return settings.folder_id;
}

void SetShowScreenShot(BOOL val)
{
    settings.show_screenshot = val;
}

BOOL GetShowScreenShot(void)
{
    return settings.show_screenshot;
}

void SetShowFolderList(BOOL val)
{
    settings.show_folderlist = val;
}

BOOL GetShowFolderList(void)
{
    return settings.show_folderlist;
}

void SetShowStatusBar(BOOL val)
{
    settings.show_statusbar = val;
}

BOOL GetShowStatusBar(void)
{
    return settings.show_statusbar;
}

void SetShowToolBar(BOOL val)
{
    settings.show_toolbar = val;
}

BOOL GetShowToolBar(void)
{
    return settings.show_toolbar;
}

void SetShowPictType(int val)
{
    settings.show_pict_type = val;
}

int GetShowPictType(void)
{
    return settings.show_pict_type;
}

void SetDefaultGame(const char *name)
{
    strcpy(settings.default_game,name);
}

const char *GetDefaultGame(void)
{
    return settings.default_game;
}

void SetWindowArea(AREA *area)
{
    memcpy(&settings.area, area, sizeof(AREA));
}

void GetWindowArea(AREA *area)
{
    memcpy(area, &settings.area, sizeof(AREA));
}

void SetListFont(LOGFONT *font)
{
    memcpy(&settings.list_font, font, sizeof(LOGFONT));
}

void GetListFont(LOGFONT *font)
{
    memcpy(font, &settings.list_font, sizeof(LOGFONT));
}

void SetListFontColor(COLORREF uColor)
{
    if (settings.list_font_color == GetSysColor(COLOR_WINDOWTEXT))
        settings.list_font_color = (COLORREF)-1;
    else
        settings.list_font_color = uColor;
}

COLORREF GetListFontColor(void)
{
    if (settings.list_font_color == (COLORREF)-1)
        return (GetSysColor(COLOR_WINDOWTEXT));

    return settings.list_font_color;
}

void SetColumnWidths(int width[])
{
    int i;

    for (i=0; i < COLUMN_MAX; i++)
        settings.column_width[i] = width[i];
}

void GetColumnWidths(int width[])
{
    int i;

    for (i=0; i < COLUMN_MAX; i++)
        width[i] = settings.column_width[i];
}

void SetSplitterPos(int splitterId, int pos)
{
    if (splitterId < SPLITTER_MAX)
        settings.splitter[splitterId] = pos;
}

int  GetSplitterPos(int splitterId)
{
    if (splitterId < SPLITTER_MAX)
        return settings.splitter[splitterId];

    return -1; /* Error */
};

void SetColumnOrder(int order[])
{
    int i;

    for (i = 0; i < COLUMN_MAX; i++)
        settings.column_order[i] = order[i];
}

void GetColumnOrder(int order[])
{
    int i;

    for (i = 0; i < COLUMN_MAX; i++)
        order[i] = settings.column_order[i];
}

void SetColumnShown(int shown[])
{
    int i;

    for (i = 0; i < COLUMN_MAX; i++)
        settings.column_shown[i] = shown[i];
}

void GetColumnShown(int shown[])
{
    int i;

    for (i = 0; i < COLUMN_MAX; i++)
        shown[i] = settings.column_shown[i];
}

void SetSortColumn(int column)
{
    settings.sort_reverse = (column < 0) ? TRUE : FALSE;
    settings.sort_column = abs(column) - 1;
}

int GetSortColumn(void)
{
    int column = settings.sort_column + 1;

    if (settings.sort_reverse)
        column = -(column);

    return column;
}

const char* GetRomDirs(void)
{
    return settings.romdirs;
}

void SetRomDirs(const char* paths)
{
    if (settings.romdirs != NULL)
    {
        MyFree(settings.romdirs);
        settings.romdirs = NULL;
    }

    if (paths != NULL)
        settings.romdirs = strdup(paths);
}

const char* GetCfgDir(void)
{
    return settings.cfgdir;
}

void SetCfgDir(const char* path)
{
    if (settings.cfgdir != NULL)
    {
        MyFree(settings.cfgdir);
        settings.cfgdir = NULL;
    }

    if (path != NULL)
        settings.cfgdir = strdup(path);
}

const char* GetNvramDir(void)
{
    return settings.nvramdir;
}

void SetNvramDir(const char* path)
{
    if (settings.nvramdir != NULL)
    {
        MyFree(settings.nvramdir);
        settings.nvramdir = NULL;
    }

    if (path != NULL)
        settings.nvramdir = strdup(path);
}

const char* GetInpDir(void)
{
    return settings.inpdir;
}

void SetInpDir(const char* path)
{
    if (settings.inpdir != NULL)
    {
        MyFree(settings.inpdir);
        settings.inpdir = NULL;
    }

    if (path != NULL)
        settings.inpdir = strdup(path);
}

const char* GetImgDir(void)
{
    return settings.snapdir;
}

void SetImgDir(const char* path)
{
    if (settings.snapdir != NULL)
    {
        MyFree(settings.snapdir);
        settings.snapdir = NULL;
    }

    if (path != NULL)
        settings.snapdir = strdup(path);
}

const char* GetIconDir(void)
{
    return settings.icondir;
}

void SetIconDir(const char* path)
{
    if (settings.icondir != NULL)
    {
        MyFree(settings.icondir);
        settings.icondir = NULL;
    }

    if (path != NULL)
        settings.icondir = strdup(path);
}

const char* GetStateDir(void)
{
    return settings.statedir;
}

void SetStateDir(const char* path)
{
    if (settings.statedir != NULL)
    {
        MyFree(settings.statedir);
        settings.statedir = NULL;
    }

    if (path != NULL)
        settings.statedir = strdup(path);
}

const char* GetMemcardDir(void)
{
    return settings.memcarddir;
}

void SetMemcardDir(const char* path)
{
    if (settings.memcarddir != NULL)
    {
        MyFree(settings.memcarddir);
        settings.memcarddir = NULL;
    }

    if (path != NULL)
        settings.memcarddir = strdup(path);
}

const char* GetFlyerDir(void)
{
    return settings.flyerdir;
}

void SetFlyerDir(const char* path)
{
    if (settings.flyerdir != NULL)
    {
        MyFree(settings.flyerdir);
        settings.flyerdir = NULL;
    }

    if (path != NULL)
        settings.flyerdir = strdup(path);
}

const char* GetCabinetDir(void)
{
    return settings.cabinetdir;
}

void SetCabinetDir(const char* path)
{
    if (settings.cabinetdir != NULL)
    {
        MyFree(settings.cabinetdir);
        settings.cabinetdir = NULL;
    }

    if (path != NULL)
        settings.cabinetdir = strdup(path);
}

const char* GetMarqueeDir(void)
{
    return settings.marqueedir;
}

void SetMarqueeDir(const char* path)
{
    if (settings.marqueedir != NULL)
    {
        MyFree(settings.marqueedir);
        settings.marqueedir = NULL;
    }

    if (path != NULL)
        settings.marqueedir = strdup(path);
}

const char* GetComboDir(void)
{
    return settings.combodir;
}

void SetComboDir(const char* path)
{
    if (settings.combodir != NULL)
    {
        MyFree(settings.combodir);
        settings.combodir = NULL;
    }

    if (path != NULL)
        settings.combodir = strdup(path);
}

const char* GetWavDir(void)
{
    return settings.wavdir;
}

void SetWavDir(const char* path)
{
    if (settings.wavdir != NULL)
    {
        MyFree(settings.wavdir);
        settings.wavdir = NULL;
    }

    if (path != NULL)
        settings.wavdir = strdup(path);
}

void ResetGameOptions(int num_game)
{
    int play_count;
    int has_roms;

    assert(0 <= num_game && num_game < num_games);

    play_count  = game[num_game].play_count;
    has_roms    = game[num_game].has_roms;

    game[num_game]              = global;
    game[num_game].use_default  = TRUE;
    game[num_game].play_count   = play_count;
    game[num_game].has_roms     = has_roms;
}

void ResetGameDefaults(void)
{
    bResetGameDefs = TRUE;
}

void ResetAllGameOptions(void)
{
    int i;

    for (i = 0; i < num_games; i++)
        ResetGameOptions(i);
}

int  GetHasRoms(int num_game)
{
    assert(0 <= num_game && num_game < num_games);

    return game[num_game].has_roms;
}

void SetHasRoms(int num_game, int has_roms)
{
    assert(0 <= num_game && num_game < num_games);

    game[num_game].has_roms = has_roms;
}

int  GetIsFavorite(int num_game)
{
    assert(0 <= num_game && num_game < num_games);

    return game[num_game].is_favorite;
}

void SetIsFavorite(int num_game, int is_favorite)
{
    assert(0 <= num_game && num_game < num_games);

    game[num_game].is_favorite = is_favorite;
}

void IncrementPlayCount(int num_game)
{
    assert(0 <= num_game && num_game < num_games);

    game[num_game].play_count++;
}

void SetFolderFlags(char *folderName, DWORD dwFlags)
{
    // Save play_count to the registry
    SaveFolderFlags(folderName, dwFlags);
}

int GetPlayCount(int num_game)
{
    assert(0 <= num_game && num_game < num_games);

    return game[num_game].play_count;
}

/***************************************************************************
    Internal functions
 ***************************************************************************/

static void LoadOptions(void)
{
    char temp[80], filename[80];
    char *linebuf;
    FILE *fp;

    linebuf = MyMalloc(LINEBUF_SIZE);
    if (linebuf == NULL)
        return;

    fp = fopen(DEFAULT_CFG, "r");
    if (fp == NULL)
    {
        MyFree(linebuf);
        return;
    }

    CfgReset();

    if (CfgReadDefaultFile(fp, linebuf) == TRUE)
    {
        LoadCfgSettingsOptions(temp);

        if (!bResetGUI && settings.version_check)
        {
            if (strcmp(temp, GetVersionString()) != 0)
            {
                char msg[400];
				char oldInfoMsg[400];

				connect_strings(oldInfoMsg, WIN32_options_oldinfo01, WIN32_options_oldinfo05);

                sprintf(msg, oldInfoMsg, temp, build_version);
                if (MessageBox(0, msg, win32_getstring(WIN32_options_oldinfotitle), MB_YESNO | MB_ICONQUESTION) == IDYES)
                {
                    bResetGUI = TRUE;
                    bResetGameDefs = TRUE;
                }
            }
        }

        if (bResetGUI)
        {
            FILE *fp2;

			sprintf(filename,"%s/%s",GetCfgDir(),GUIBACK_CFG);
			fp2 = fopen(filename, "r");

            if (fp2 != NULL)
            {
                CfgReset();
                if (CfgReadDefaultFile(fp2, linebuf) == TRUE)
                    LoadCfgSettingsOptions(temp);
                fclose(fp2);
            }
        }

        if (settings.list_font_color != -1)
        {
            if (settings.list_font_color == GetSysColor(COLOR_WINDOWTEXT))
                settings.list_font_color = (COLORREF)-1;
        }

        LoadCfgGameOptions(&global);

    }

	sprintf(filename,"%s/%s",GetCfgDir(),GAMES_CFG);
    fp = fopen(filename, "r");
    if (fp == NULL)
    {
        MyFree(linebuf);
        return;
    }

    CfgReadFile(fp, linebuf);

    fclose(fp);

    MyFree(linebuf);
}


static void SaveOptions(void)
{
    FILE  *fp;
	char  filename[80];
    int   i;

    fp = fopen(DEFAULT_CFG, "w");
    if (fp == NULL)
        return;

    CfgReset();

    SaveCfgSettingsOptions();

    global.use_default = FALSE;
    SaveCfgGameOptions(&global);

    CfgWriteFile(fp, "default");

    fclose(fp);

	sprintf(filename,"%s/%s",GetCfgDir(),GAMES_CFG);
    fp = fopen(filename, "w");
    if (fp == NULL)
        return;

    for (i = 0; i < num_games; i++)
    {
        CfgReset();
        SaveCfgGameOptions(&game[i]);
        CfgWriteFile(fp, drivers[i]->name);
    }

    fclose(fp);
}


static void SaveBackupOptions(void)
{
    FILE  *fp;
	char  filename[80];

	sprintf(filename,"%s/%s",GetCfgDir(),GUIBACK_CFG);
    fp = fopen(filename, "w");
    if (fp == NULL)
        return;

    CfgReset();
    SaveCfgSettingsOptions();
    CfgWriteFile(fp, "default");

    fclose(fp);

	sprintf(filename,"%s/%s",GetCfgDir(),GAMEBACK_CFG);
    fp = fopen(filename, "w");
    if (fp == NULL)
        return;

    global.use_default = FALSE;

    CfgReset();
    SaveCfgGameOptions(&global);
    CfgWriteFile(fp, "default");

    fclose(fp);
}


DWORD GetFolderFlags(char *folderName)
{
    FILE  *fp;
    char  *linebuf;
    char  value[16], filename[80];
    DWORD size;
    DWORD flags = 0;

	sprintf(filename,"%s/%s",GetCfgDir(),FOLDERS_CFG);
    fp = fopen(filename, "r");
    if (fp != NULL)
    {
        linebuf = MyMalloc(LINEBUF_SIZE);
        if (linebuf != NULL)
        {
            CfgReset();
            CfgReadFile(fp, linebuf);
            MyFree(linebuf);
        }
        if (CfgQueryValue(folderName, NULL, &size) == ERROR_SUCCESS)
        {
            if (CfgQueryValue(folderName, value, &size) == ERROR_SUCCESS)
                sscanf(value, "%08lX", &flags);
        }
        fclose(fp);
    }

    return flags;
}

static void SaveFolderFlags(char *folderName, DWORD dwFlags)
{
    FILE *fp;
    char  *linebuf;
    char  value[16], filename[80];
    int   i, count;

    sprintf(value, "%08lX", dwFlags);

    CfgReset();

	sprintf(filename,"%s/%s",GetCfgDir(),FOLDERS_CFG);
    fp = fopen(filename, "r");
    if (fp != NULL)
    {
        linebuf = MyMalloc(LINEBUF_SIZE);
        if (linebuf != NULL)
        {
            CfgReadFile(fp, linebuf);
            MyFree(linebuf);
        }
        fclose(fp);
    }

    fp = fopen(filename, "w");
    if (fp != NULL)
    {
        CfgSetValue(folderName, value, strlen(value));
        CfgWriteFile(fp, "folders");
        fclose(fp);
    }

    count = 0;
    for (i = 0; cfg[i].value; i++)
        count += atoi(cfg[i].value);

    if (count == 0)
        remove(filename);
}

void SaveGameOptions(int game_num)
{
    FILE  *fp;
	char  filename[80];
    int   i;

	sprintf(filename,"%s/%s",GetCfgDir(),GAMES_CFG);
    fp = fopen(filename, "w");
    if (fp == NULL)
        return;

    for (i = 0; i < num_games; i++)
    {
        CfgReset();
        SaveCfgGameOptions(&game[i]);
        CfgWriteFile(fp, drivers[i]->name);
    }

    fclose(fp);
}

void SaveDefaultOptions(void)
{
    FILE  *fp;

    fp = fopen(DEFAULT_CFG, "w");
    if (fp == NULL)
        return;

    CfgReset();

    SaveCfgSettingsOptions();

    global.use_default = FALSE;
    SaveCfgGameOptions(&global);

    CfgWriteFile(fp, "default");

    fclose(fp);
}

static void SaveCfgSettingsOptions(void)
{
    int i;

    PutCfgStringOption("SaveVersion", GetVersionString());
    PutCfgBoolOption("VersionCheck", settings.version_check);
    PutCfgOption("FontColor", settings.list_font_color);
    PutCfgBoolOption("ResetGUI", bResetGUI);
    PutCfgBoolOption("ResetGameDefaults", bResetGameDefs);

    for (i = 0; i < NUM_SETTINGS; i++)
        PutCfgObj(&cfgSettings[i]);
}

static void LoadCfgSettingsOptions(char *szVersion)
{
    int i;

    strcpy(szVersion, GetCfgStringOption("SaveVersion"));
    settings.version_check   = GetCfgBoolOption("VersionCheck");
    bResetGUI                = GetCfgBoolOption("ResetGUI");
    bResetGameDefs           = GetCfgBoolOption("ResetGameDefaults");
    settings.list_font_color = GetCfgOption("FontColor");

    for (i = 0; i < NUM_SETTINGS; i++)
        GetCfgObj(&cfgSettings[i]);
}

static void SaveCfgGameOptions(options_type *o)
{
    int i;

    if (o->use_default == TRUE)
        CfgReset();

    PutCfgOption("PlayCount", o->play_count);
    PutCfgOption("HasRoms", o->has_roms);
    PutCfgBoolOption("Favorite", o->is_favorite);

    if (o->use_default == TRUE)
        return;

    gOpts = *o;

    for (i = 0; i < NUM_GAMEOPTS; i++)
        PutCfgObj(&cfgGameOpts[i]);
}

static void LoadCfgGameOptions(options_type *o)
{
    int   i;
    DWORD value;
    DWORD size;

    if ((value = GetCfgOption("PlayCount")) != -1)
        o->play_count = value;
    if ((value = GetCfgOption("HasRoms")) != -1)
        o->has_roms = value;
    o->is_favorite = GetCfgBoolOption("Favorite");

    if (CfgQueryValue("Monitor", &value, &size) != ERROR_SUCCESS)
       return;

    o->use_default = FALSE;

    gOpts = *o;

    for (i = 0; i < NUM_GAMEOPTS; i++)
        GetCfgObj(&cfgGameOpts[i]);

    /* copy options back out */
    *o = gOpts;
}

static DWORD GetCfgOption(char *name)
{
    char *ptr;
    DWORD value = -1;

    ptr = GetCfgStringOption(name);
    if (ptr != NULL)
        value = atoi(ptr);

    return value;
}

static BOOL GetCfgBoolOption(char *name)
{
    char *ptr;
    BOOL value = FALSE;

    ptr = GetCfgStringOption(name);
    if (ptr != NULL)
        value = (*ptr == '0') ? FALSE : TRUE;

    return value;
}

static char *GetCfgStringOption(char *name)
{
    DWORD dwSize;
    static char str[LINEBUF_SIZE];

    if (CfgQueryValue(name, NULL, &dwSize) != ERROR_SUCCESS)
        return NULL;

    CfgQueryValue(name, str, &dwSize);

    return str;
}

static void PutCfgOption(char *name, DWORD value)
{
    char str[16];

    sprintf(str, "%ld", value);

    CfgSetValue(name, (void *)&str, strlen(str));
}

static void PutCfgBoolOption(char *name, BOOL value)
{
    char str[2];

    str[0] = (value) ? '1' : '0';
    str[1] = '\0';

    CfgSetValue(name, (void *)str, 2);
}

static void PutCfgStringOption(char *name, char *option)
{
    CfgSetValue(name, (void *)option, strlen(option));
}

static void GetCfgObj(CFG_OPTIONS *cfgOpts)
{
    char    *cName = cfgOpts->m_cName;
    char    *pString;
	char    *token;
    double  *pDouble;
    struct stat stat_buffer;

    switch(cfgOpts->m_iType)
    {
    case CO_DOUBLE:
        pDouble = (double *)cfgOpts->m_vpData;
        if ((pString = GetCfgStringOption(cName)) != NULL)
            sscanf(pString, "%lf", pDouble);
        break;

    case CO_BOOL:
        *(BOOL *)cfgOpts->m_vpData = GetCfgBoolOption(cName);
        break;

    case CO_INT:
        *(int *)cfgOpts->m_vpData = GetCfgOption(cName);
        break;

    case CO_STRING:
        if ((pString = GetCfgStringOption(cName)) != NULL)
            strcpy((char *)cfgOpts->m_vpData, pString);
        break;

    case CO_PSTRING:
        if ((pString = GetCfgStringOption(cName)) != NULL)
        {
            if (*(char **)cfgOpts->m_vpData != NULL)
                MyFree(*(char **)cfgOpts->m_vpData);
            *(char **)cfgOpts->m_vpData = strdup(pString);
        }
	    token = strtok(pString, ";");
		while (token != NULL)				// tH_Ȃꍇ͍쐬B
		{
			if (stat(token, &stat_buffer) != 0)
				mkdir(token);
		    token = strtok(NULL, ";");
		}
        break;

    case CO_ENCODE:
        if ((pString = GetCfgStringOption(cName)) != NULL)
            cfgOpts->decode(pString, cfgOpts->m_vpData);
        break;

    default:
        break;
    }
}

static void PutCfgObj(CFG_OPTIONS *cfgOpt)
{
    char    *pString;
    double  *pDouble;
    char    *cName = cfgOpt->m_cName;
    char    cTemp[80];

    switch (cfgOpt->m_iType)
    {
    case CO_DOUBLE:
        pDouble = (double *)cfgOpt->m_vpData;
        sprintf(cTemp, "%03.02f", *pDouble);
        PutCfgStringOption(cName, cTemp);
        break;

    case CO_BOOL:
        PutCfgBoolOption(cName, *(BOOL *)cfgOpt->m_vpData);
        break;

    case CO_INT:
        PutCfgOption(cName, *(int *)cfgOpt->m_vpData);
        break;

    case CO_STRING:
        pString = (char *)cfgOpt->m_vpData;
        if (pString)
            PutCfgStringOption(cName, pString);
        break;

    case CO_PSTRING:
        pString = *(char **)cfgOpt->m_vpData;
        if (pString)
            PutCfgStringOption(cName, pString);
        break;

    case CO_ENCODE:
        cfgOpt->encode(cfgOpt->m_vpData, cTemp);
        PutCfgStringOption(cName, cTemp);
        break;

    default:
        break;
    }
}

/*********************************************************************/

static void ColumnEncodeString(void *data, char *str)
{
    int* value = (int*)data;
    int  i;
    char tmpStr[100];

    sprintf(tmpStr, "%d", value[0]);

    strcpy(str, tmpStr);

    for (i = 1; i < COLUMN_MAX; i++)
    {
        sprintf(tmpStr, ",%d", value[i]);
        strcat(str, tmpStr);
    }
}

static void ColumnDecodeString(const char *str, void *data)
{
    int* value = (int*)data;
    int  i;
    char *s, *p;
    char tmpStr[100];

    if (str == NULL)
        return;

    strcpy(tmpStr, str);
    p = tmpStr;
    
    for (i = 0; p && i < COLUMN_MAX; i++)
    {
        s = p;
        
        if ((p = strchr(s,',')) != NULL && *p == ',')
        {
            *p = '\0';
            p++;
        }
        value[i] = atoi(s);
    }
}

static void ColumnDecodeWidths(const char *str, void *data)
{
    if (settings.view == VIEW_REPORT)
        ColumnDecodeString(str, data);
}

static void SplitterEncodeString(void *data, char *str)
{
    int* value = (int*)data;
    int  i;
    char tmpStr[100];

    sprintf(tmpStr, "%d", value[0]);
    
    strcpy(str, tmpStr);

    for (i = 1; i < SPLITTER_MAX; i++)
    {
        sprintf(tmpStr, ",%d", value[i]);
        strcat(str, tmpStr);
    }
}

static void SplitterDecodeString(const char *str, void *data)
{
    int* value = (int*)data;
    int  i;
    char *s, *p;
    char tmpStr[100];

    if (str == NULL)
        return;

    strcpy(tmpStr, str);
    p = tmpStr;
    
    for (i = 0; p && i < SPLITTER_MAX; i++)
    {
        s = p;
        
        if ((p = strchr(s,',')) != NULL && *p == ',')
        {
            *p = '\0';
            p++;
        }
        value[i] = atoi(s);
    }
}

static void ListDecodeString(const char *str, void *data)
{
    int* value = (int*)data;
    int i;

    *value = VIEW_REPORT;

    for (i = VIEW_LARGE_ICONS; i < VIEW_MAX; i++)
    {
        if (strcmp(str, view_modes[i]) == 0)
        {
            *value = i;
            return;
        }
    }
}

static void ListEncodeString(void *data, char *str)
{
    int* value = (int*)data;

    strcpy(str, view_modes[*value]);
}

/* Parse the given comma-delimited string into a LOGFONT structure */
static void FontDecodeString(const char *str, void *data)
{
    LOGFONT* f = (LOGFONT*)data;
    char *   ptr;
    int      n[8];

    sscanf(str, "%ld,%ld,%ld,%ld,%ld,%i,%i,%i,%i,%i,%i,%i,%i",
           &f->lfHeight,
           &f->lfWidth,
           &f->lfEscapement,
           &f->lfOrientation,
           &f->lfWeight,
#if 0
           &f->lfItalic,
           &f->lfUnderline,
           &f->lfStrikeOut,
           &f->lfCharSet,
           &f->lfOutPrecision,
           &f->lfClipPrecision,
           &f->lfQuality,
           &f->lfPitchAndFamily);
#else
           &n[0], &n[1], &n[2], &n[3],
           &n[4], &n[5], &n[6], &n[7]);

           f->lfItalic         = (BYTE)n[0];
           f->lfUnderline      = (BYTE)n[1];
           f->lfStrikeOut      = (BYTE)n[2];
           f->lfCharSet        = (BYTE)n[3];
           f->lfOutPrecision   = (BYTE)n[4];
           f->lfClipPrecision  = (BYTE)n[5];
           f->lfQuality        = (BYTE)n[6];
           f->lfPitchAndFamily = (BYTE)n[7];
#endif
    ptr = strrchr(str, ',');
    if (ptr != NULL)
        strcpy(f->lfFaceName, ptr + 1);
}

/* Encode the given LOGFONT structure into a comma-delimited string */
static void FontEncodeString(void *data, char *str)
{
    LOGFONT* f = (LOGFONT*)data;

    sprintf(str, "%ld,%ld,%ld,%ld,%ld,%i,%i,%i,%i,%i,%i,%i,%i,%s",
            f->lfHeight,
            f->lfWidth,
            f->lfEscapement,
            f->lfOrientation,
            f->lfWeight,
            (int)f->lfItalic,
            (int)f->lfUnderline,
            (int)f->lfStrikeOut,
            (int)f->lfCharSet,
            (int)f->lfOutPrecision,
            (int)f->lfClipPrecision,
            (int)f->lfQuality,
            (int)f->lfPitchAndFamily,
            f->lfFaceName);
}

/***************************************************************************/

static BOOL CfgCreate(void)
{
    int i;

    cfg = (cfg_struct *)MyMalloc(sizeof(cfg_struct) * CFG_MAX_ENTRIES);
    if (cfg == NULL)
        return FALSE;

    for (i = 0; i < CFG_MAX_ENTRIES; i++)
    {
        cfg[i].name  = NULL;
        cfg[i].value = NULL;
    }

    return TRUE;
}

static void CfgDestroy(void)
{
    int i;

    if (cfg == NULL)
        return;

    for (i = 0; i < CFG_MAX_ENTRIES; i++)
    {
        if (cfg[i].name)  MyFree(cfg[i].name);
        if (cfg[i].value) MyFree(cfg[i].value);
    }

    MyFree(cfg);
    cfg = NULL;
}

static void CfgReset(void)
{
    int i;

    if (cfg == NULL)
        return;

    for (i = 0; i < CFG_MAX_ENTRIES; i++)
    {
        if (cfg[i].name)  MyFree(cfg[i].name);
        if (cfg[i].value) MyFree(cfg[i].value);

        cfg[i].name  = NULL;
        cfg[i].value = NULL;
    }
}

static BOOL CfgReadDefaultFile(FILE *fp, char *linebuf)
{
    char *name;
    char *value;
    BOOL found = FALSE;

    while (fgets(linebuf, LINEBUF_SIZE, fp))
    {
        if (linebuf[0] == '[')
        {
            char *p = strchr(linebuf, ']');
            if (p != NULL) *p = '\0';
            if (strcmp("default", &linebuf[1]) == 0)
            {
                found = TRUE;
                break;
            }
        }
    }

    if (found == FALSE)
        return FALSE;

    while (fgets(linebuf, LINEBUF_SIZE, fp))
    {
        if (linebuf[0] == '[')
            break;

        if (linebuf[0] == '#')
            continue;

        name = strtok(linebuf, " \t\r\n");
        if (name == NULL)
            continue;

        value = strtok(&linebuf[24], "\t\r\n");

		if (value == NULL)	// ʐMo[W󔒂̏ꍇ
		{
			CfgSetValue_ex(name);
		}
		else
		{
			CfgSetValue(name, value, strlen(value));
		}
    }

    return TRUE;
}


static void CfgReadFile(FILE *fp, char *linebuf)
{
    char *name;
    char *value;
    char *p;
    int  index = -1;

    while (fgets(linebuf, LINEBUF_SIZE, fp))
    {
        if (linebuf[0] == '[')
        {
            if (index != -1)
                LoadCfgGameOptions(&game[index]);

            p = strchr(linebuf, ']');
            if (p == NULL)
                continue;

            *p = '\0';
            index = GetIndexByName(&linebuf[1]);
            CfgReset();
            continue;
        }

        if (linebuf[0] == '#')
            continue;

        name = strtok(linebuf, " \t\r\n");
        if (name == NULL)
            continue;

        value = strtok(&linebuf[24], "\t\r\n");

		if (value == NULL)	// ʐMo[W󔒂̏ꍇ
		{
			CfgSetValue_ex(name);
		}
		else
		{
			CfgSetValue(name, value, strlen(value));
		}
    }

    if (index != -1)
        LoadCfgGameOptions(&game[index]);
}


static void CfgWriteFile(FILE *fp, const char *keyname)
{
    int   i, j;
    DWORD size;
    char  saved[CFG_MAX_ENTRIES];

    memset(saved, 0, CFG_MAX_ENTRIES);

    fprintf(fp, "[%s]\n", keyname);

    if (CfgQueryValue("Monitor", NULL, &size) == ERROR_SUCCESS)
    {
        for (i = 0; i < NUM_GAMEOPTS; i++)
        {
            for (j = 0; cfg[j].name; j++)
            {
                if (saved[j]) continue;

                if (strcmp(cfgGameOpts[i].m_cName, cfg[j].name) == 0)
                {
                    fprintf(fp, "%-21s   %s\n", cfg[j].name, cfg[j].value);
                    saved[j] = 1;
                }
            }
        }
    }

    if (CfgQueryValue("rompath", NULL, &size) == ERROR_SUCCESS)
    {
        for (i = 0; i < NUM_SETTINGS; i++)
        {
            for (j = 0; cfg[j].name; j++)
            {
                if (saved[j]) continue;

                if (strcmp(cfgSettings[i].m_cName, cfg[j].name) == 0)
                {
                    fprintf(fp, "%-21s   %s\n", cfg[j].name, cfg[j].value);
                    saved[j] = 1;
                }
            }
        }
    }

    for (i = 0; cfg[i].name; i++)
    {
        if (!saved[i])
            fprintf(fp, "%-21s   %s\n", cfg[i].name, cfg[i].value);
    }

    fprintf(fp, "\n");
}


static long CfgQueryValue(const char *name, void *value, DWORD *size)
{
    int i;

    for (i = 0; cfg[i].name; i++)
    {
        if (strcmp(name, cfg[i].name) == 0)
        {
            if (value == NULL)
            {
                *size = strlen(cfg[i].value);
                return ERROR_SUCCESS;
            }
            else
            {
                if (cfg[i].value)
                {
                    strcpy(value, cfg[i].value);
                    return ERROR_SUCCESS;
                }
                else
                    return !ERROR_SUCCESS;
            }
        }
    }

    return !ERROR_SUCCESS;
}


static void CfgSetValue(const char *name, void *value, DWORD size)
{
    int i;
    BOOL found = FALSE;

    for (i = 0; cfg[i].name; i++)
    {
        if (strcmp(name, cfg[i].name) == 0)
        {
            found = TRUE;
            break;
        }
    }

    if (!found)
    {
        cfg[i].name = MyMalloc(strlen(name) + 1);
        strcpy(cfg[i].name, name);
    }

    cfg[i].value = MyRealloc(cfg[i].value, size + 1);
    strcpy(cfg[i].value, value);
}


static void CfgSetValue_ex(const char *name)
{
    int i;
    BOOL found = FALSE;

    for (i = 0; cfg[i].name; i++)
    {
        if (strcmp(name, cfg[i].name) == 0)
        {
            found = TRUE;
            break;
        }
    }

    if (!found)
    {
        cfg[i].name = MyMalloc(strlen(name) + 1);
        strcpy(cfg[i].name, name);
    }

    cfg[i].value = MyRealloc(cfg[i].value, 2);
    strcpy(cfg[i].value, "\0");
}


INLINE int GetIndexByName(const char *name)
{
    int i;

    for (i = 0; i < num_games; i++)
    {
        if (strcmp(name, drivers[i]->name) == 0)
            return i;
    }
    return -1;
}

#ifndef BUILD_ENGLISH
BOOL UseJapaneseList(void)
{
    return settings.use_japanese_list;
}

void SetUseJapaneseList(BOOL bUseJPN)
{
    settings.use_japanese_list = bUseJPN;
}
#endif

BOOL UseModifyThe(void)
{
    return settings.modify_the;
}

void SetUseModifyThe(BOOL bUseThe)
{
    settings.modify_the = bUseThe;
}

#ifdef KAILLERA
BOOL GetUseImeInChat(void)
{
    return settings.use_ime_in_chat;
}

void SetUseImeInChat(BOOL bUseIME)
{
    settings.use_ime_in_chat = bUseIME;
}

BOOL GetUseFixedForm(void)
{
    return settings.use_fixed_form;
}

void SetUseFixedForm(BOOL bUseFF)
{
    settings.use_fixed_form = bUseFF;
}
#endif

void GetVersionName(char *bVersionName)
{
	strcpy(bVersionName, settings.version_name);
}

void SetVersionName(char *bVersionName)
{
	strcpy(settings.version_name, bVersionName);
}

char * GetVersionString(void)
{
    static char tmp[120];

#if defined(BETA_VERSION)
    sprintf(tmp,"%d BETA %d - %s", VERSION, BETA_VERSION, __DATE__);
#else
    sprintf(tmp,"%d - %s", VERSION, __DATE__);
#endif
    return tmp;
}

/* End of options.c */
