#ifndef DRIVER_H
#define DRIVER_H

#include "osd_cpu.h"
#include "memory.h"
#include "mamedbg.h"
#include "osdepend.h"
#include "mame.h"
#include "common.h"
#include "drawgfx.h"
#include "palette.h"
#include "cpuintrf.h"
#include "sndintrf.h"
#include "input.h"
#include "inptport.h"
#include "usrintrf.h"
#include "cheat.h"
#include "profiler.h"
#include "ui_font.h"


#define MAX_CPU 2	/* MAX_CPU is the maximum number of CPUs which cpuintrf.c */
					/* can run at the same time. Currently, 8 is enough. */


#define MAX_SOUND 1	/* MAX_SOUND is the maximum number of sound subsystems */
					/* which can run at the same time. Currently, 5 is enough. */



struct MachineDriver
{
	/* basic machine hardware */
	struct MachineCPU cpu[MAX_CPU];
	float frames_per_second;
	int vblank_duration;	/* in microseconds - see description below */
	int cpu_slices_per_frame;	/* for multicpu games. 1 is the minimum, meaning */
								/* that each CPU runs for the whole video frame */
								/* before giving control to the others. The higher */
								/* this setting, the more closely CPUs are interleaved */
								/* and therefore the more accurate the emulation is. */
								/* However, an higher setting also means slower */
								/* performance. */
	void (*init_machine)(void);

    /* video hardware */
	int screen_width,screen_height;
	const struct rectangle default_visible_area;	/* the visible area can be changed at */
									/* run time, but it should never be larger than the */
									/* one specified here, in order not to force the */
									/* OS dependant code to resize the display window. */
	const struct GfxDecodeInfo *gfxdecodeinfo;
	unsigned int total_colors;	/* palette is 3*total_colors bytes long */
	unsigned int color_table_len;	/* length in shorts of the color lookup table */
	void (*vh_init_palette)(unsigned char *palette, unsigned short *colortable,const unsigned char *color_prom);

	int video_attributes;	/* ASG 081897 */

	void (*vh_eof_callback)(void);	/* called every frame after osd_update_video_and_audio() */
									/* This is useful when there are operations that need */
									/* to be performed every frame regardless of frameskip, */
									/* e.g. sprite buffering or collision detection. */
	int (*vh_start)(void);
	void (*vh_stop)(void);
	void (*vh_update)(struct osd_bitmap *bitmap,int full_refresh);

	/* sound hardware */
	int sound_attributes;
	int obsolete1;
	int obsolete2;
	int obsolete3;
	struct MachineSound sound[MAX_SOUND];

	/*
	   use this to manage nvram/eeprom/cmos/etc.
	   It is called before the emulation starts and after it ends. Note that it is
	   NOT called when the game is reset, since it is not needed.
	   file == 0, read_or_write == 0 -> first time the game is run, initialize nvram
	   file != 0, read_or_write == 0 -> load nvram from disk
	   file == 0, read_or_write != 0 -> not allowed
	   file != 0, read_or_write != 0 -> save nvram to disk
	 */
	void (*nvram_handler)(void *file,int read_or_write);
};



/* VBlank is the period when the video beam is outside of the visible area and */
/* returns from the bottom to the top of the screen to prepare for a new video frame. */
/* VBlank duration is an important factor in how the game renders itself. MAME */
/* generates the vblank_interrupt, lets the game run for vblank_duration microseconds, */
/* and then updates the screen. This faithfully reproduces the behaviour of the real */
/* hardware. In many cases, the game does video related operations both in its vblank */
/* interrupt, and in the normal game code; it is therefore important to set up */
/* vblank_duration accurately to have everything properly in sync. An example of this */
/* is Commando: if you set vblank_duration to 0, therefore redrawing the screen BEFORE */
/* the vblank interrupt is executed, sprites will be misaligned when the screen scrolls. */

/* Here are some predefined, TOTALLY ARBITRARY values for vblank_duration, which should */
/* be OK for most cases. I have NO IDEA how accurate they are compared to the real */
/* hardware, they could be completely wrong. */
#define DEFAULT_60HZ_VBLANK_DURATION 0
#define DEFAULT_30HZ_VBLANK_DURATION 0
/* If you use IPT_VBLANK, you need a duration different from 0. */
#define DEFAULT_REAL_60HZ_VBLANK_DURATION 2500
#define DEFAULT_REAL_30HZ_VBLANK_DURATION 2500



/* flags for video_attributes */

/* bit 0 of the video attributes indicates raster or vector video hardware */
#define	VIDEO_TYPE_RASTER			0x0000

/* bit 2 of the video attributes indicates whether or not the driver modifies the palette */
#define	VIDEO_MODIFIES_PALETTE	0x0004

/* flags for sound_attributes */
#define	SOUND_SUPPORTS_STEREO		0x0001

struct GameDriver
{
	const char *source_file;	/* set this to __FILE__ */
	const struct GameDriver *clone_of;	/* if this is a clone, point to */
										/* the main version of the game */
	const char *name;
#ifndef BUILD_ENGLISH
	const char *descriptione;
#endif
	const char *description;
	const char *sortname;
	const char *year;
	const char *manufacturer;
	const int  genre;
	const int  sound_type;
	const UINT16 sound_bank[8];
	const struct MachineDriver *drv;
	const struct InputPortTiny *input_ports;
	void (*driver_init)(void);	/* optional function to be called during initialization */
								/* This is called ONCE, unlike Machine->init_machine */
								/* which is called every time the game is reset. */

	const struct RomModule *rom;

	UINT32 flags;	/* see defines below */
};

/* values for the flags field */

#define GAME_NOT_WORKING			0x0008
#define GAME_WRONG_COLORS			0x0010	/* colors are totally wrong */
#define GAME_IMPERFECT_COLORS		0x0020	/* colors are not 100% accurate, but close */
#define GAME_NO_SOUND				0x0040	/* sound is missing */
#define GAME_IMPERFECT_SOUND		0x0080	/* sound is known to be wrong */
#define	GAME_REQUIRES_16BIT			0x0100	/* cannot fit in 256 colors */
#define GAME_NO_COCKTAIL			0x0200	/* screen flip support is missing */
#define GAME_UNEMULATED_PROTECTION	0x0400	/* game's protection not fully emulated */
#define NOT_A_DRIVER				0x4000	/* set by the fake "root" driver_ and by "containers" */
											/* e.g. driver_neogeo. */
#define GAME_NEOGEO_DRIVER			0x0000
#define GAME_RASTER_DRIVER			0x8000
#define GAME_neogeo_DRIVER			0x0000
#define GAME_raster_DRIVER			0x8000

enum {
	SOUND_UNKNOWN = 0,
	SOUND_SNK,
	SOUND_ADK,
	SOUND_DATAEAST,
	SOUND_FACE,
	SOUND_NAZKA,
	SOUND_SUMMY,
	SOUND_TECHNOS,
	SOUND_TECMO,
	SOUND_YUMEKOBO,
	SOUNDTYPE_MAX
};

enum {
	OTHER = 0,
	ACTION,
	SHOOT,
	FIGHT,
	RACE,
	SPORTS,
	PUZZLE,
	TABLE,
	QUIZ,
	GENRE_MAX
};

#ifdef BUILD_ENGLISH
#define GAME(YEAR,NAME,PARENT,MACHINE,INPUT,INIT,GENRE,SOUND_TYPE,BANK1,BANK2,BANK3,BANK4,BANK5,BANK6,BANK7,BANK8,COMPANY,FULLNAME)	\
extern struct GameDriver driver_##PARENT;	\
struct GameDriver driver_##NAME =			\
{											\
	__FILE__,								\
	&driver_##PARENT,						\
	#NAME,									\
	FULLNAME,								\
	FULLNAME,								\
	#YEAR,									\
	COMPANY,								\
	GENRE,									\
	SOUND_TYPE,								\
	{ BANK1, BANK2, BANK3, BANK4, BANK5, BANK6, BANK7, BANK8 }, \
	&machine_driver_neogeo,					\
	input_ports_##INPUT,					\
	init_##INIT,							\
	rom_##NAME,								\
	GAME_REQUIRES_16BIT | GAME_##MACHINE##_DRIVER	\
};

#define GAMEX(YEAR,NAME,PARENT,MACHINE,INPUT,INIT,GENRE,COMPANY,FULLNAME,FLAGS)	\
extern struct GameDriver driver_##PARENT;	\
struct GameDriver driver_##NAME =			\
{											\
	__FILE__,								\
	&driver_##PARENT,						\
	#NAME,									\
	FULLNAME,								\
	FULLNAME,								\
	#YEAR,									\
	COMPANY,								\
	GENRE,									\
	0,										\
	{ 0, 0, 0, 0, 0, 0, 0, 0 },				\
	&machine_driver_neogeo,					\
	input_ports_##INPUT,					\
	init_##INIT,							\
	rom_##NAME,								\
	GAME_REQUIRES_16BIT | FLAGS | GAME_##MACHINE##_DRIVER	\
};
#else
#define GAME(YEAR,NAME,PARENT,MACHINE,INPUT,INIT,GENRE,SOUND_TYPE,BANK1,BANK2,BANK3,BANK4,BANK5,BANK6,BANK7,BANK8,COMPANY,ENGNAME,FULLNAME,SORTNAME)	\
extern struct GameDriver driver_##PARENT;	\
struct GameDriver driver_##NAME =			\
{											\
	__FILE__,								\
	&driver_##PARENT,						\
	#NAME,									\
	ENGNAME,								\
	FULLNAME,								\
	SORTNAME,								\
	#YEAR,									\
	COMPANY,								\
	GENRE,									\
	SOUND_TYPE,								\
	{ BANK1, BANK2, BANK3, BANK4, BANK5, BANK6, BANK7, BANK8 }, \
	&machine_driver_neogeo,					\
	input_ports_##INPUT,					\
	init_##INIT,							\
	rom_##NAME,								\
	GAME_REQUIRES_16BIT | GAME_##MACHINE##_DRIVER	\
};

#define GAMEX(YEAR,NAME,PARENT,MACHINE,INPUT,INIT,GENRE,COMPANY,ENGNAME,FULLNAME,SORTNAME,FLAGS)	\
extern struct GameDriver driver_##PARENT;	\
struct GameDriver driver_##NAME =			\
{											\
	__FILE__,								\
	&driver_##PARENT,						\
	#NAME,									\
	ENGNAME,								\
	FULLNAME,								\
	SORTNAME,								\
	#YEAR,									\
	COMPANY,								\
	GENRE,									\
	0,										\
	{ 0, 0, 0, 0, 0, 0, 0, 0 },				\
	&machine_driver_neogeo,					\
	input_ports_##INPUT,					\
	init_##INIT,							\
	rom_##NAME,								\
	GAME_REQUIRES_16BIT | FLAGS | GAME_##MACHINE##_DRIVER	\
};
#endif

/* this allows to leave the INIT field empty in the GAME() macro call */
#define init_0 0

/* Fppbgԍ̃}N */
#define TRANS_COLOR	4096

extern struct GameDriver *drivers[];

#endif
