/***************************************************************************

  vidhrdw.c

  Functions to emulate the video hardware of the machine.

Important!	There are two types of NeoGeo romdump - MVS & MGD2.  They are both
converted to a standard format in the vh_start routines.


Graphics information:

0x00000 - 0xdfff	: Blocks of sprite data, each 0x80 bytes:
	Each 0x80 block is made up of 0x20 double words, their format is:
	Word: Sprite number (16 bits)
	Byte: Palette number (8 bits)
	Byte: Bit 0: X flip
		  Bit 1: Y flip
		  Bit 2: Automatic animation flag (4 tiles?)
		  Bit 3: Automatic animation flag (8 tiles?)
		  Bit 4: MSB of sprite number (confirmed, Karnov_r, Mslug). See note.
		  Bit 5: MSB of sprite number (MSlug2)
		  Bit 6: MSB of sprite number (Kof97)
		  Bit 7: Unknown for now

	Each double word sprite is drawn directly underneath the previous one,
	based on the starting coordinates.

0x0e000 - 0x0ea00	: Front plane fix tiles (8*8), 2 bytes each

0x10000: Control for sprites banks, arranged in words

	Bit 0 to 3 - Y zoom LSB
	Bit 4 to 7 - Y zoom MSB (ie, 1 byte for Y zoom).
	Bit 8 to 11 - X zoom, 0xf is full size (no scale).
	Bit 12 to 15 - Unknown, probably unused

0x10400: Control for sprite banks, arranged in words

	Bit 0 to 5: Number of sprites in this bank (see note below).
	Bit 6 - If set, this bank is placed to right of previous bank (same Y-coord).
	Bit 7 to 15 - Y position for sprite bank.

0x10800: Control for sprite banks, arranged in words
	Bit 0 to 5: Unknown
	Bit 7 to 15 - X position for sprite bank.

Notes:

* If rom set has less than 0x10000 tiles then msb of tile must be ignored
(see Magician Lord).

***************************************************************************/

#pragma warning( disable : 4799 ) // Disables EMMS warning for inline assembly

#include "driver.h"
#include "common.h"
#include "usrintrf.h"
#include "drivers/neogeo.h"

//#define MMX_SPRITE   1
//#define MMX_FIXCHAR  1
//#define MMX_CLEAR    1
//#define MMX_MARGEBMP 1
#if (MMX_SPRITE || MMX_FIXCHAR || MMX_CLEAR)
#define NEOGEO_MMX
#endif

//#define USE_RGBTABLE

#define PALETTE_SIZE  0x2000

static data16_t *neogeo_vidram16;
static data16_t *neogeo_paletteram16;	/* pointer to 1 of the 2 palette banks */
static data16_t *neogeo_palettebank[2]; /* 0x100*16 2 byte palette entries */
static int neogeo_palette_index;
static data16_t neogeo_vidram16_modulo;
static data16_t neogeo_vidram16_offset;
static int high_tile;
static int vhigh_tile;
static int vvhigh_tile;
static int palette_swap_pending;
static int fix_bank;
static int object_number;

extern data16_t *neogeo_ram16;
extern unsigned int neogeo_frame_counter;
extern int neogeo_machine_country;

static UINT32 *l_y_skip;
static UINT32 dda_y_skip[0x100][0x21][17];
static UINT8 yskip[0x100][0x21];
static UINT8 mmy[0x100][0x10];

static struct osd_bitmap *sub_bitmap = 0;
static int raster_start_line = 240;

int no_of_tiles;
int gururin_fix = 0;
int sdodgeb_fix = 0;
int garou_fix = 0;

#ifdef USE_RGBTABLE
static UINT8 RGB[32768][3];
#endif

void alt_backcolor_exp(void);

static int code_data = 0;

/******************************************************************************/

#ifdef USE_RGBTABLE
static void build_rgb_table(void)
{
	int color, r, g, b;

	for (color = 0; color < 32768; color++)
	{
		r = ((color >> 7) & 0x1e) | ((color >> 14) & 0x01);
		g = ((color >> 3) & 0x1e) | ((color >> 13) & 0x01);
		b = ((color << 1) & 0x1e) | ((color >> 12) & 0x01);

		r = (r << 3) | (r >> 2);
		g = (g << 3) | (g >> 2);
		b = (b << 3) | (b >> 2);

		RGB[color][0] = r;
		RGB[color][1] = g;
		RGB[color][2] = b;
	}
}
#endif

static void build_zoom_table(void)
{
	if (options.disable_nglo)
	{
		int i, my, zy, dday;

		logerror ("\n");
		logerror ("no use ng-lo.rom\n");

		for (zy = 0; zy <= 0xff; zy++)
		{
			dday = 0;

			for (my = 0; my < 0x21; my++)
			{
				if (zy != 0xff)
				{
					yskip[zy][my] = 0;
					dda_y_skip[zy][my][0] = 0;

					for(i = 0; i < 16; i++)
					{
						dda_y_skip[zy][my][i + 1] = 0;
						dday -= zy + 1;
						if (dday <= 0)
						{
							dday += 256;
							yskip[zy][my]++;
						}
						dda_y_skip[zy][my][yskip[zy][my]]++;
					}
				}
				else
				{
					dda_y_skip[zy][my][0] = 0;

					for(i = 1; i < 16; i++)
						dda_y_skip[zy][my][i] = 1;

					yskip[zy][my] = 16;
				}

				if (my < 0x10)
				{
					mmy[zy][my] = ((my * 16 * 256) / (zy + 1) + 15) / 16;
					if (mmy[zy][my] > 0x10)
						mmy[zy][my] = 0x10;
				}
			}
		}
	}
	else
	{
		int i, j, my, zy, line;
		int tile1, tile0, yoffs1, yoffs0;
		UINT8 *zoomy_rom;

		logerror ("\n");
		logerror ("use ng-lo.rom\n");

		for (zy = 0; zy <= 0xff; zy++)
		{
			zoomy_rom = memory_region(REGION_GFX3) + (zy << 8);
			line = 0;

			for (i = 0; i < 0x10; i++)
			{
				yskip[zy][i] = 0;
				dda_y_skip[zy][i][0] = 0x10;

				for (j = 1; j < 0x10; j++)
				{
					dda_y_skip[zy][i][j] = 0;
				}
			}

			for (i = 0; i < zy + 1; i++)
			{
				yoffs1 = zoomy_rom[i] & 0x0f;
				tile1 = zoomy_rom[i] >> 4;

				if (line == 0)
				{
					dda_y_skip[zy][tile1][0] = yoffs1;
					yskip[zy][tile1]++;
					line = 1;
				}
				else
				{
					yoffs0 = zoomy_rom[i - 1] & 0x0f;
					tile0 = zoomy_rom[i - 1] >> 4;

					if (tile1 == tile0)
					{
						dda_y_skip[zy][tile1][line] = yoffs1 - yoffs0;
						yskip[zy][tile1]++;
						line++;
					}
					else
					{
						dda_y_skip[zy][tile0][line] = 0x10 - yoffs0;
						dda_y_skip[zy][tile1][0] = yoffs1;
						yskip[zy][tile1]++;
						line = 1;
					}
				}

				if (i == zy)
				{
					dda_y_skip[zy][tile1][line] = 0x10 - yoffs1;
				}
			}

			for (my = 0; my < 0x10; my++)
			{
				if (zoomy_rom[my * 0x10] == 0xff)
					mmy[zy][my] = 0x10;
				else
					mmy[zy][my] = zoomy_rom[my * 0x10] >> 4;
			}
		}

		for (zy = 0; zy <= 0xff; zy++)
		{
			for (my = 0x10; my <= 0x20; my++)
			{
				yskip[zy][my] = yskip[zy][my - 0x10];

				for (line = 0; line <= yskip[zy][my]; line++)
				{
					dda_y_skip[zy][my][line] = dda_y_skip[zy][my - 0x10][line];
				}
			}
		}
	}

#if 0	// test code
	{
		int zy, my, i;

		for (zy = 0; zy <= 0xff; zy++)
		{
			for (my = 0; my < 0x21; my++)
			{
				logerror("yskip[%d][%d] = %2X\n", zy, my, yskip[zy][my]);
				logerror("mmy[%d][%d] = %2X\n", zy, my, mmy[zy][my]);
				for(i = 0; i < 16; i++)
				{
					logerror("dda_y_skip[%d][%d][%d] = %2X\n", zy, my, i, dda_y_skip[zy][my][i]);
				}
			}
		}
	}
#endif	// test code
}

/******************************************************************************/

void neogeo_vh_stop(void)
{
	if (neogeo_palettebank[0])
		MyFree(neogeo_palettebank[0]);
	neogeo_palettebank[0] = NULL;

	if (neogeo_palettebank[1])
		MyFree(neogeo_palettebank[1]);
	neogeo_palettebank[1] = NULL;

	if (neogeo_vidram16)
		MyFree(neogeo_vidram16);
	neogeo_vidram16 = NULL;

	if (neogeo_ram16)
		MyFree(neogeo_ram16);
	neogeo_ram16 = NULL;

	if (sub_bitmap)
		osd_free_bitmap(sub_bitmap);
	sub_bitmap = NULL;
}

static int common_vh_start(void)
{
	neogeo_palettebank[0] = NULL;
	neogeo_palettebank[1] = NULL;
	neogeo_vidram16 = NULL;

	neogeo_palettebank[0] = MyMalloc(PALETTE_SIZE);
	if (!neogeo_palettebank[0])
	{
		neogeo_vh_stop();
		return 1;
	}

	neogeo_palettebank[1] = MyMalloc(PALETTE_SIZE);
	if (!neogeo_palettebank[1])
	{
		neogeo_vh_stop();
		return 1;
	}

	/* 0x20000 bytes even though only 0x10c00 is used */
	neogeo_vidram16 = MyMalloc(0x20000);
	if (!neogeo_vidram16)
	{
		neogeo_vh_stop();
		return 1;
	}

	sub_bitmap = osd_alloc_bitmap(40*8, 32*8);
	if (!sub_bitmap)
	{
		neogeo_vh_stop();
		return 1;
	}

	memset(neogeo_palettebank[0], 0, PALETTE_SIZE);
	memset(neogeo_palettebank[1], 0, PALETTE_SIZE);
	memset(neogeo_vidram16,       0, 0x20000);

	neogeo_paletteram16 = neogeo_palettebank[0];
	neogeo_palette_index = 0;
	neogeo_vidram16_modulo = 1;
	neogeo_vidram16_offset = 0;
	fix_bank = 0;
	palette_swap_pending = 0;
	raster_start_line    = 240;

	build_zoom_table();

#ifdef USE_RGBTABLE
	build_rgb_table();
#endif

	return 0;
}


INLINE void decodetile(int tileno)
{
	unsigned char swap[128];
	UINT32 *gfxdata;
	int x,y;
	unsigned int pen;


	gfxdata = (UINT32 *)&memory_region(REGION_GFX2)[128 * tileno];

	memcpy(swap, gfxdata, 128);

	for (y = 0; y < 16; y++)
	{
		UINT32 dw;

		dw = 0;
		for (x = 0; x < 8; x++)
		{
			pen  = ((swap[64 + 4*y + 3] >> x) & 1) << 3;
			pen |= ((swap[64 + 4*y + 1] >> x) & 1) << 2;
			pen |= ((swap[64 + 4*y + 2] >> x) & 1) << 1;
			pen |=  (swap[64 + 4*y    ] >> x) & 1;
			dw |= pen << 4*x;
			Machine->gfx[2]->pen_usage[tileno] |= (1 << pen);
		}
		*(gfxdata++) = dw;

		dw = 0;
		for (x = 0; x < 8; x++)
		{
			pen  = ((swap[4*y + 3] >> x) & 1) << 3;
			pen |= ((swap[4*y + 1] >> x) & 1) << 2;
			pen |= ((swap[4*y + 2] >> x) & 1) << 1;
			pen |=  (swap[4*y    ] >> x) & 1;
			dw |= pen << 4*x;
			Machine->gfx[2]->pen_usage[tileno] |= (1 << pen);
		}
		*(gfxdata++) = dw;
	}
}

int neogeo_mvs_vh_start(void)
{
	no_of_tiles = memory_region_length(REGION_GFX2) / 128;

	high_tile   = (no_of_tiles > 0x10000);
	vhigh_tile  = (no_of_tiles > 0x20000);
	vvhigh_tile = (no_of_tiles > 0x40000);

	Machine->gfx[2]->total_elements = no_of_tiles;

	if (Machine->gfx[2]->pen_usage)
		MyFree(Machine->gfx[2]->pen_usage);
	Machine->gfx[2]->pen_usage = MyMalloc(no_of_tiles * sizeof(int));
	memset(Machine->gfx[2]->pen_usage, 0, no_of_tiles * sizeof(int));

	/* tiles are not decoded yet. They will be decoded later as they are used. */
	/* pen_usage is used as a marker of decoded tiles: if it is 0, then the tile */
	/* hasn't been decoded yet. */

	return common_vh_start();
}

/******************************************************************************/

INLINE void swap_palettes(void)
{
	int i;

	for (i = 0; i < 0x2000 >> 1; i++)
	{
		data16_t newword = neogeo_paletteram16[i];
#ifndef USE_RGBTABLE
		int r, g, b;

		r = ((newword >> 7) & 0x1e) | ((newword >> 14) & 0x01);
		g = ((newword >> 3) & 0x1e) | ((newword >> 13) & 0x01);
		b = ((newword << 1) & 0x1e) | ((newword >> 12) & 0x01);

		r = (r << 3) | (r >> 2);
		g = (g << 3) | (g >> 2);
		b = (b << 3) | (b >> 2);

		palette_change_color(i, r, g, b);
#else
		palette_change_color(i, RGB[newword][0], RGB[newword][1], RGB[newword][2]);
#endif
		alt_backcolor_exp();
	}

	palette_swap_pending = 0;
}

static void neogeo_setpalbank(int n)
{
	if (neogeo_palette_index != n)
	{
		neogeo_palette_index = n;
		neogeo_paletteram16 = neogeo_palettebank[n];
		palette_swap_pending = 1;
	}
}

WRITE16_HANDLER( neogeo_setpalbank0_16_w )
{
	neogeo_setpalbank(0);
}

WRITE16_HANDLER( neogeo_setpalbank1_16_w )
{
	neogeo_setpalbank(1);
}

READ16_HANDLER( neogeo_paletteram16_r )
{
	return neogeo_paletteram16[offset];
}

WRITE16_HANDLER( neogeo_paletteram16_w )
{
	data16_t oldword, newword;
#ifndef USE_RGBTABLE
	int r,g,b;
#endif

	oldword = newword = neogeo_paletteram16[offset];
	COMBINE_DATA(&newword);

	if (oldword == newword)
		return;

	neogeo_paletteram16[offset] = newword;

#ifndef USE_RGBTABLE
	r = ((newword >> 7) & 0x1e) | ((newword >> 14) & 0x01);
	g = ((newword >> 3) & 0x1e) | ((newword >> 13) & 0x01);
	b = ((newword << 1) & 0x1e) | ((newword >> 12) & 0x01);

	r = (r << 3) | (r >> 2);
	g = (g << 3) | (g >> 2);
	b = (b << 3) | (b >> 2);

	palette_change_color(offset, r, g, b);
#else
	palette_change_color(offset, RGB[newword][0], RGB[newword][1], RGB[newword][2]);
#endif
	alt_backcolor_exp();
}

/******************************************************************************/

WRITE16_HANDLER( neogeo_vidram16_offset_w )
{
	COMBINE_DATA(&neogeo_vidram16_offset);
}

READ16_HANDLER( neogeo_vidram16_data_r )
{
	return neogeo_vidram16[neogeo_vidram16_offset];
}

WRITE16_HANDLER( neogeo_vidram16_data_w )
{
	COMBINE_DATA(&neogeo_vidram16[neogeo_vidram16_offset]);
	neogeo_vidram16_offset += neogeo_vidram16_modulo;
	neogeo_vidram16_offset &= 0xffff;
}

/* Modulo can become negative , Puzzle Bobble Super Sidekicks and a lot */
/* of other games use this */
WRITE16_HANDLER( neogeo_vidram16_modulo_w )
{
	COMBINE_DATA(&neogeo_vidram16_modulo);
}

READ16_HANDLER( neogeo_vidram16_modulo_r )
{
	return neogeo_vidram16_modulo;
}

WRITE16_HANDLER( neo_board_fix_16_w )
{
	fix_bank = 1;
}

WRITE16_HANDLER( neo_game_fix_16_w )
{
	fix_bank = 0;
}

#ifdef MMX_SPRITE

/******************************************************************************
   MMX helper macros
******************************************************************************/

//#define NEW_MACRO

const static UINT32 PALMASK = 0x0f0f0f0f;

#define MMXSpriteStart()                  \
	__asm mov       esi, fspr             \
	__asm mov       edi, bm               \

#define MMXSpriteLoopStart()              \
	__asm mov       edx, l_y_skip         \
	__asm mov       eax, l                \
	__asm mov       eax, [edx + eax * 4]  \
	__asm imul      dy                    \
	__asm lea       esi, [esi + eax]

#define MMXSpriteLoopEnd(LABEL)           \
	__asm inc       l                     \
	__asm dec       zy                    \
	__asm lea       edi, [edi + 336 * 2]  \
	__asm jnz       LABEL

/*----------------------------------------*/

#define MMXSpriteIncrementLine()          \
	__asm inc       l                     \
	__asm lea       edi, [edi + 336 * 2]

#define MMXSpriteSetColorMask()           \
	__asm movd      mm5, color            \
	__asm psllw     mm5, 4                \
	__asm punpcklwd mm5, mm5              \
	__asm punpckldq mm5, mm5

#define MMXSpriteSetZeroMask()            \
	__asm pxor      mm6, mm6

#define MMXSpriteSetPaletteMask()         \
	__asm movd      mm7, PALMASK          \
	__asm punpckldq mm7, mm7

#define MMXSpriteDecode(src)              \
	__asm movd      mm0, src              \
	__asm punpcklbw mm0, mm6              \
	__asm movq      mm1, mm0              \
	__asm psllw     mm1, 4                \
	__asm por       mm0, mm1              \
	__asm pand      mm0, mm7

#define MMXSpriteDecodeFlipX(src)         \
	__asm movd      mm0, src              \
	__asm punpcklbw mm0, mm6              \
	__asm movq      mm1, mm0              \
	__asm psllw     mm0, 8                \
	__asm psrlw     mm1, 4                \
	__asm por       mm0, mm1              \
	__asm pand      mm0, mm7

#define MMXSpriteUnpackLowWord()          \
	__asm punpcklbw mm0, mm6              \
	__asm por       mm0, mm5

#define MMXSpriteUnpackHighWord()         \
	__asm movq      mm1, mm0              \
	__asm punpckhbw mm1, mm6              \
	__asm por       mm1, mm5

#define MMXSpriteUnpackByte(src, dst)     \
	__asm movd      dst, src              \
	__asm punpcklbw dst, mm6              \
	__asm por       dst, mm5

#define MMXSpriteDrawLow(dst)             \
	__asm movq      mm3, [edi + dst]      \
	__asm movq      mm2, mm0              \
	__asm pcmpeqw   mm2, mm5              \
	__asm pand      mm3, mm2              \
	__asm pandn     mm2, mm0              \
	__asm por       mm2, mm3              \
	__asm movq      [edi + dst], mm2

#define MMXSpriteDrawHigh(dst)            \
	__asm movq      mm3, [edi + dst]      \
	__asm movq      mm2, mm1              \
	__asm pcmpeqw   mm2, mm5              \
	__asm pand      mm3, mm2              \
	__asm pandn     mm2, mm1              \
	__asm por       mm2, mm3              \
	__asm movq      [edi + dst], mm2

#define MMXSpriteByteSwap(src, dst)       \
	__asm mov       dst, src              \
	__asm bswap     dst

#define MMXSpriteNibbleSwap(src, dst)     \
	__asm mov       dst, src              \
	__asm bswap     dst                   \
	__asm mov       ecx, dst              \
	__asm and       ecx, 00f0f0f0fh       \
	__asm shl       ecx, 4                \
	__asm and       dst, 0f0f0f0f0h       \
	__asm shr       dst, 4                \
	__asm or        dst, ecx

/*---------------------------------------------------------------------------*/

#ifdef NEW_MACRO
#define MMXSpriteSetMask()                \
	MMXSpriteSetZeroMask()                \
	MMXSpriteSetPaletteMask()
#else
#define MMXSpriteSetMask()                \
	MMXSpriteSetColorMask()               \
	MMXSpriteSetZeroMask()                \
	MMXSpriteSetPaletteMask()
#endif

#define MMXSpriteFullDraw(s, d1, d2)      \
	MMXSpriteDecode(s)                    \
	MMXSpriteUnpackHighWord()             \
	MMXSpriteUnpackLowWord()              \
	MMXSpriteDrawLow(d1)                  \
	MMXSpriteDrawHigh(d2)

#define MMXSpriteFullDrawFlipX(s, d1, d2) \
	MMXSpriteDecodeFlipX(s)               \
	MMXSpriteUnpackHighWord()             \
	MMXSpriteUnpackLowWord()              \
	MMXSpriteDrawLow(d1)                  \
	MMXSpriteDrawHigh(d2)

#define MMXSpriteHalfDraw(s, d)           \
	MMXSpriteDecode(s)                    \
	MMXSpriteUnpackLowWord()              \
	MMXSpriteDrawLow(d)


/******************************************************************************
   ͌AANZX̑ŁAMMXgpłƑ卷ȂƂ
   ނxȂĂ悤ȁB
******************************************************************************/

#ifndef NEW_MACRO

static void NeoMVSDrawGfx(struct osd_bitmap *bitmap,const struct GfxElement *gfx,
		unsigned int code,unsigned int color,int flipx,int flipy,int sx,int sy,
		int zx,int zy,const struct rectangle *clip)
{
	int oy, ey, dy;
	UINT32 *fspr;

	fspr = (UINT32 *)memory_region(REGION_GFX2);

	/* Mish/AJP - Most clipping is done in main loop */
	oy = sy;
	ey = sy + zy -1; 	/* Clip for size of zoomed object */

	if (sy < clip->min_y) sy = clip->min_y;
	if (ey >= clip->max_y) ey = clip->max_y;

	if (!flipy) /* normal */
	{
		dy = 2;
		fspr+=code*32 + (sy-oy)*2;
	}
	else /* Y flip */
	{
		dy = -2;
		fspr+=(code+1)*32 - 2 - (sy-oy)*2;
	}

	zy = ey - sy + 1;

	{
		int l = 0;
		UINT16 *bm = &bitmap->line[sy][sx];

		if (!flipx)	/* normal */
		{
			switch (zx)
			{
				case 16: // 11111111 11111111
						 // 01234567 89ABCDEF
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov   esi, fspr

						MMXSpriteFullDraw([esi], 0, 8)
						MMXSpriteFullDraw([esi + 4], 16, 24)
						MMXSpriteIncrementLine()
					}
					return;

				case 15: // 1111111x 11111111
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm {
							mov   esi, fspr
							mov   eax, [esi]
							mov   edx, [esi + 4]
							mov   ecx, edx
							and   eax, 00fffffffh
							and   ecx, 00000000fh
							shl   ecx, 28
							or    eax, ecx
							shr   edx, 4
						}

						MMXSpriteFullDraw(eax, 0, 8)
						MMXSpriteFullDraw(edx, 16, 24)
						MMXSpriteIncrementLine()
					}
					return;

				case 14: // 1111111x 1111111x
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm {
							mov   esi, fspr
							mov   eax, [esi]
							mov   edx, [esi + 4]
							mov   ecx, edx
							and   eax, 00fffffffh
							and   ecx, 00000000fh
							shl   ecx, 28
							or    eax, ecx
							and   edx, 00ffffff0h
							shr   edx, 4
						}

						MMXSpriteFullDraw(eax, 0, 8)
						MMXSpriteFullDraw(edx, 16, 24)
						MMXSpriteIncrementLine()
					}
					return;

				case 13: // 11x1111x 1111x111
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm {
							mov   esi, fspr
							mov   eax, [esi]
							mov   edx, [esi + 4]
							mov   ecx, eax
							and   eax, 00ffff000h
							shr   eax, 4
							and   ecx, 0000000ffh
							or    eax, ecx
							mov   ecx, edx
							and   ecx, 0000000ffh
							shl   ecx, 24
							or    eax, ecx
							mov   ecx, edx
							and   ecx, 00000ff00h
							shr   ecx, 8
							and   edx, 0fff00000h
							shr   edx, 12
							or    edx, ecx
						}

						MMXSpriteFullDraw(eax, 0, 8)
						MMXSpriteFullDraw(edx, 16, 24)
						MMXSpriteIncrementLine()
					}
					return;


				case 12: // 111x111x 111x111x
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm {
							mov   esi, fspr
							mov   eax, [esi]
							mov   edx, [esi + 4]
							mov   ecx, eax
							and   eax, 00fff0000h
							shr   eax, 4
							and   ecx, 000000fffh
							or    eax, ecx
							mov   ecx, edx
							and   ecx, 0000000ffh
							shl   ecx, 24
							or    eax, ecx
							mov   ecx, edx
							and   ecx, 00fff0000h
							shr   ecx, 12
							and   edx, 000000f00h
							shr   edx, 8
							or    edx, ecx
						}

						MMXSpriteFullDraw(eax, 0, 8)
						MMXSpriteHalfDraw(edx, 16)
						MMXSpriteIncrementLine()
					}
					return;

				case 11: // x11x11x1 11x11x11
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm {
							mov   esi, fspr
							mov   eax, [esi]
							mov   edx, [esi + 4]
							mov   ecx, eax
							and   ecx, 000000ff0h
							shr   ecx, 4
							mov   ebx, eax
							and   ebx, 000ff0000h
							shr   ebx, 8
							or    ecx, ebx
							and   eax, 0f0000000h
							shr   eax, 12
							or    eax, ecx
							mov   ecx, edx
							and   ecx, 0000000ffh
							shl   ecx, 20
							or    eax, ecx
							mov   ecx, edx
							and   ecx, 00000f000h
							shl   ecx, 16
							or    eax, ecx
							mov   ecx, edx
							and   ecx, 0000f0000h
							shr   ecx, 16
							and   edx, 0ff000000h
							shr   edx, 20
							or    edx, ecx
						}

						MMXSpriteFullDraw(eax, 0, 8)
						MMXSpriteHalfDraw(edx, 16)
						MMXSpriteIncrementLine()
					}
					return;

				case 10: // 1x1x1x11 1x1x1x11
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm {
							mov       esi, fspr
							mov       eax, [esi]
							mov       ecx, eax
							and       eax, 00f0f0f0fh
							and       ecx, 0f0000000h
							shr       ecx, 28
							mov       edx, [esi + 4]
							mov       ebx, edx
							and       edx, 0000f0f0fh
							shl       edx, 8
							or        edx, ecx
							mov       ecx, ebx
							and       ecx, 00f000000h
							shr       ecx, 24
							and       ebx, 0f0000000h
							shr       ebx, 12
							or        ecx, ebx
						}

						MMXSpriteUnpackByte(eax, mm0)
						MMXSpriteUnpackByte(edx, mm1)
						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)

						__asm {
							movd      mm0, ecx
							por       mm0, mm5
						}

						MMXSpriteDrawLow(16)
						MMXSpriteIncrementLine()
					}
					return;

				case 9: // 1x1x1x1x 1x1x1x11
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi]
							and       eax, 00f0f0f0fh
							mov       edx, [esi + 4]
							mov       ecx, edx
							and       edx, 00f0f0f0fh
							and       ecx, 0f0000000h
							shr       ecx, 28
						}

						MMXSpriteUnpackByte(eax, mm0)
						MMXSpriteUnpackByte(edx, mm1)
						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)

						__asm {
							movd      mm0, ecx
							por       mm0, mm5
						}

						MMXSpriteDrawLow(16)
						MMXSpriteIncrementLine()
					}
					return;

				case 8: // 1x1x1x1x 1x1x1x1x
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi]
							and       eax, 00f0f0f0fh
							mov       edx, [esi + 4]
							and       edx, 00f0f0f0fh
						}

						MMXSpriteUnpackByte(eax, mm0)
						MMXSpriteUnpackByte(edx, mm1)
						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)
						MMXSpriteIncrementLine()
					}
					return;

				case 7: // x1x1x1xx 1x1x1x1x
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi]
							and       eax, 000f0f0f0h
							shr       eax, 4
							mov       edx, [esi + 4]
							mov       ecx, edx
							and       ecx, 00000000fh
							shl       ecx, 24
							or        eax, ecx
							and       edx, 00f0f0f00h
							shr       edx, 8
						}

						MMXSpriteUnpackByte(eax, mm0)
						MMXSpriteUnpackByte(edx, mm1)
						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)
						MMXSpriteIncrementLine()
					}
					return;

				case 6: // x1xx1xx1 1xx1xx1x
					MMXSpriteSetColorMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi]
							mov       edx, eax
							mov       ecx, eax
							and       eax, 0000000f0h
							shr       eax, 4
							and       edx, 0000f0000h
							or        eax, edx
							and       ecx, 0f0000000h
							shr       ecx, 28
							mov       edx, [esi + 4]
							mov       ebx, edx
							and       ebx, 00000000fh
							shl       ebx, 16
							or        ecx, ebx
							mov       ebx, edx
							and       ebx, 00000f000h
							shr       ebx, 12
							and       edx, 00f000000h
							shr       edx, 8
							or        edx, ebx
							movd      mm0, eax
							movd      mm1, ecx
;							punpckldq mm0, mm1
							psllq     mm1, 32
							por       mm0, mm1
							por       mm0, mm5
							movd      mm1, edx
							por       mm1, mm5
						}

						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)
						MMXSpriteIncrementLine()
					}
					return;

				case 5: // xx1xx1xx 1xx1xx1x
					MMXSpriteSetColorMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi]
							mov       edx, eax
							and       eax, 000000f00h
							shr       eax, 8
							and       edx, 000f00000h
							shr       edx, 4
							or        eax, edx
							mov       edx, [esi + 4]
							mov       ecx, edx
							mov       ebx, edx
							and       ecx, 00000000fh
							and       ebx, 00000f000h
							shl       ebx, 4
							or        ecx, ebx
							and       edx, 00f000000h
							shr       edx, 24
							movd      mm0, eax
							movd      mm1, ecx
;							punpckldq mm0, mm1
							psllq     mm1, 32
							por       mm0, mm1
							por       mm0, mm5
							movd      mm1, edx
							por       mm1, mm5
						}

						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)
						MMXSpriteIncrementLine()
					}
					return;

				case 4: // 1xxx1xxx 1xxx1xxx
					MMXSpriteSetColorMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi]
							mov       edx, [esi + 4]
							and       eax, 0000f000fh
							and       edx, 0000f000fh
							movd      mm0, eax
							movd      mm1, edx
							psllq     mm1, 32
							por       mm0, mm1
							por       mm0, mm5
						}

						MMXSpriteDrawLow(0)
						MMXSpriteIncrementLine()
					}
					return;

				case 3: // xxx1xxxx 1xxxx1xx
					MMXSpriteSetColorMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi]
							and       eax, 00000f000h
							shr       eax, 12
							mov       edx, [esi + 4]
							mov       ecx, edx
							and       ecx, 00000000fh
							shl       ecx, 16
							or        eax, ecx
							and       edx, 000f00000h
							shr       edx, 20
							movd      mm0, eax
							movd      mm1, edx
							punpckldq mm0, mm1
							por       mm0, mm5
						}

						MMXSpriteDrawLow(0)
						MMXSpriteIncrementLine()
					}
					return;

				case 2: // 1xxxxxxx 1xxxxxxx
					MMXSpriteSetColorMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi]
							and       eax, 00000000fh
							mov       edx, [esi + 4]
							and       edx, 00000000fh
							shl       edx, 16
							or        eax, edx
							movd      mm0, eax
							por       mm0, mm5
						}

						MMXSpriteDrawLow(0)
						MMXSpriteIncrementLine()
					}
					return;

				case 1: // 1xxxxxxx xxxxxxxx
					MMXSpriteSetColorMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi]
							and       eax, 00000000fh
							movd      mm0, eax
							por       mm0, mm5
						}

						MMXSpriteDrawLow(0)
						MMXSpriteIncrementLine()
					}
					return;
			}
		}
		else	/* X flip */
		{
			switch (zx)
			{
				case 16: // 11111111 11111111
						 // FEDCBA98 76543210
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteByteSwap([esi + 4], eax)
						MMXSpriteByteSwap([esi], edx)
						MMXSpriteFullDrawFlipX(eax, 0, 8)
						MMXSpriteFullDrawFlipX(edx, 16, 24)
						MMXSpriteIncrementLine()
					}
					return;

				case 15: // 11111111 x1111111
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteNibbleSwap([esi + 4], eax)
						MMXSpriteNibbleSwap([esi], edx)

						__asm shr       edx, 4

						MMXSpriteFullDraw(eax, 0, 8)
						MMXSpriteFullDraw(edx, 16, 24)
						MMXSpriteIncrementLine()
					}
					return;

				case 14: // x1111111 x1111111
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteNibbleSwap([esi + 4], eax)
						MMXSpriteNibbleSwap([esi], edx)

						__asm {
							mov       ecx, edx
							and       edx, 00ffffff0h
							shr       edx, 4
							shr       eax, 4
							and       ecx, 0000000f0h
							shl       ecx, 24
							or        eax, ecx
						}

						MMXSpriteFullDraw(eax, 0, 8)
						MMXSpriteFullDraw(edx, 16, 24)
						MMXSpriteIncrementLine()
					}
					return;

				case 13: // 111x1111 x1111x11
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteNibbleSwap([esi + 4], eax)
						MMXSpriteNibbleSwap([esi], edx)

						__asm {
							mov       ecx, edx
							mov       ebx, edx
							and       ecx, 0ff000000h
							shr       ecx, 12
							and       edx, 0000fff00h
							shr       edx, 8
							or        edx, ecx
							and       ebx, 0000000f0h
							shl       ebx, 24
							mov       ecx, eax
							and       ecx, 0ffff0000h
							shr       ecx, 4
							and       eax, 000000fffh
							or        eax, ecx
							or        eax, ebx
						}

						MMXSpriteFullDraw(eax, 0, 8)
						MMXSpriteFullDraw(edx, 16, 24)
						MMXSpriteIncrementLine()
					}
					return;

				case 12: // x111x111 x111x111
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteNibbleSwap([esi + 4], eax)
						MMXSpriteNibbleSwap([esi], edx)

						__asm {
							mov       ecx, edx
							mov       ebx, edx
							and       ecx, 0fff00000h
							shr       ecx, 16
							and       edx, 00000f000h
							shr       edx, 12
							or        edx, ecx
							and       ebx, 000000ff0h
							shl       ebx, 20
							mov       ecx, eax
							and       ecx, 0fff00000h
							shr       ecx, 8
							and       eax, 00000fff0h
							shr       eax, 4
							or        eax, ecx
							or        eax, ebx
						}

						MMXSpriteFullDraw(eax, 0, 8)
						MMXSpriteHalfDraw(edx, 16)
						MMXSpriteIncrementLine()
					}
					return;

				case 11: // 11x11x11 1x11x11x
					MMXSpriteSetMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteNibbleSwap([esi + 4], eax)
						MMXSpriteNibbleSwap([esi], edx)

						__asm {
							mov       ecx, edx
							mov       ebx, edx
							and       ecx, 00ff00000h
							shr       ecx, 16
							and       edx, 00000f000h
							shr       edx, 12
							or        edx, ecx
							mov       ecx, ebx
							and       ecx, 000000f00h
							shl       ecx, 20
							and       ecx, 00000000fh
							shl       ebx, 24
							or        ebx, ecx
							mov       ecx, eax
							and       ecx, 0ff000000h
							shr       ecx, 8
							or        ebx, ecx
							mov       ecx, eax
							and       ecx, 0000ff000h
							shr       ecx, 4
							or        ebx, ecx
							and       eax, 0000000ffh
							or        eax, ebx
						}

						MMXSpriteFullDraw(eax, 0, 8)
						MMXSpriteHalfDraw(edx, 16)
						MMXSpriteIncrementLine()
					}
					return;

				case 10: // 11x1x1x1 11x1x1x1
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteByteSwap([esi + 4], eax)
						MMXSpriteByteSwap([esi], edx)

						__asm {
							mov       ecx, eax
							mov       ebx, eax
							and       ecx, 0000f0f0fh
							shl       ecx, 8
							and       eax, 0000000f0h
							shr       eax, 4
							or        eax, ecx
							and       ebx, 00f000000h
							shr       ebx, 24
							mov       ecx, edx
							and       edx, 0000000f0h
							shl       edx, 4
							or        edx, ebx
							mov       ebx, ecx
							and       ebx, 000000f0fh
							shl       ebx, 16
							or        edx, ebx
							mov       ebx, ecx
							and       ebx, 0000f0000h
							shr       ebx, 16
							and       ecx, 00f000000h
							shr       ecx, 8
							or        ecx, ebx
						}

						MMXSpriteUnpackByte(eax, mm0)
						MMXSpriteUnpackByte(edx, mm1)
						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)

						__asm {
							movd      mm0, ecx
							por       mm0, mm5
						}

						MMXSpriteDrawLow(16)
						MMXSpriteIncrementLine()
					}
					return;

				case 9:  // 11x1x1x1 x1x1x1x1
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteByteSwap([esi + 4], eax)
						MMXSpriteByteSwap([esi], edx)

						__asm {
							mov       ecx, eax
							mov       ebx, eax
							and       ecx, 0000f0f0fh
							shl       ecx, 8
							and       eax, 0000000f0h
							shr       eax, 4
							or        eax, ecx
							and       ebx, 00f000000h
							shr       ebx, 24
							mov       ecx, edx
							and       edx, 0000f0f0fh
							shl       edx, 8
							or        edx, ebx
							and       ecx, 00f000000h
							shr       ecx, 24
						}

						MMXSpriteUnpackByte(eax, mm0)
						MMXSpriteUnpackByte(edx, mm1)
						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)

						__asm {
							movd      mm0, ecx
							por       mm0, mm5
						}

						MMXSpriteDrawLow(16)
						MMXSpriteIncrementLine()
					}
					return;

				case 8:  // x1x1x1x1 x1x1x1x1
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteByteSwap([esi + 4], eax)
						MMXSpriteByteSwap([esi], edx)

						__asm {
							and       eax, 0f0f0f0fh
							and       edx, 0f0f0f0fh
						}

						MMXSpriteUnpackByte(eax, mm0)
						MMXSpriteUnpackByte(edx, mm1)
						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)
						MMXSpriteIncrementLine()
					}
					return;

				case 7: // x1x1x1x1 xx1x1x1x
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteByteSwap([esi + 4], eax)
						MMXSpriteByteSwap([esi], edx)

						__asm {
							and       eax, 00f0f0f0fh
							and       edx, 0f0f0f000h
							shr       edx, 12
						}

						MMXSpriteUnpackByte(eax, mm0)
						MMXSpriteUnpackByte(edx, mm1)
						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)
						MMXSpriteIncrementLine()
					}
					return;

				case 6: // x1xx1xx1 1xx1xx1x
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm {
							mov       esi, fspr
							mov       eax, [esi + 4]
							mov       edx, eax
							mov       ecx, eax
							and       eax, 00f000000h
							shr       eax, 24
							and       edx, 00000f000h
							shr       edx, 4
							or        eax, edx
							and       ecx, 00000000fh
							shl       ecx, 16
							or        eax, ecx
							mov       edx, [esi]
							mov       ebx, edx
							and       ebx, 0f0000000h
							shr       ebx, 4
							or        eax, ebx
							mov       ecx, edx
							and       edx, 0000f0000h
							shr       edx, 16
							and       ecx, 0000000f0h
							shl       ecx, 12
							or        edx, ecx
							movd      mm1, edx
							por       mm1, mm5
						}

						MMXSpriteUnpackByte(eax, mm0)
						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)
						MMXSpriteIncrementLine()
					}
					return;

				case 5: // x1xx1xx1 xx1xx1xx
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm {
							mov       esi, fspr
							mov       eax, [esi + 4]
							mov       edx, eax
							mov       ecx, eax
							and       eax, 00f000000h
							shr       eax, 24
							and       edx, 00000f000h
							shr       edx, 4
							or        eax, edx
							and       ecx, 00000000fh
							shl       ecx, 16
							or        eax, ecx
							mov       edx, [esi]
							mov       ebx, edx
							and       ebx, 000f00000h
							shl       ebx, 4
							or        eax, ebx
							and       edx, 000000f00h
							shr       edx, 8
							movd      mm1, edx
							por       mm1, mm5
						}

						MMXSpriteUnpackByte(eax, mm0)
						MMXSpriteDrawLow(0)
						MMXSpriteDrawHigh(8)
						MMXSpriteIncrementLine()
					}
					return;

				case 4: // xxx1xxx1 xxx1xxx1
					MMXSpriteSetColorMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm mov       esi, fspr

						MMXSpriteByteSwap([esi + 4], eax)
						MMXSpriteByteSwap([esi], edx)

						__asm{
							and       eax, 00f000f00h
							shr       eax, 8
							and       edx, 00f000f00h
							shr       edx, 8
							movd      mm0, eax
							movd      mm1, edx
							psllq     mm1, 32
							por       mm0, mm1
							por       mm0, mm5
						}

						MMXSpriteDrawLow(0)
						MMXSpriteIncrementLine()
					}
					return;

				case 3: // xx1xxxx1 xxxx1xxx
					MMXSpriteSetColorMask()
					MMXSpriteSetZeroMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi + 4]
							mov       edx, eax
							and       eax, 000f00000h
							shr       eax, 20
							and       edx, 00000000fh
							shl       eax, 8
							or        eax, edx
							mov       edx, [esi]
							and       edx, 00000f000h
							shl       edx, 4
							or        eax, edx
							movd      mm0, eax
							punpcklbw mm0, mm6
							por       mm0, mm5
						}

						MMXSpriteDrawLow(0)
						MMXSpriteIncrementLine()
					}
					return;

				case 2: // xxxxxxx1 xxxxxxx1
					MMXSpriteSetColorMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							mov       eax, [esi + 4]
							and       eax, 0f0000000h
							shr       eax, 28
							mov       edx, [esi]
							and       edx, 0f0000000h
							shr       edx, 12
							or        eax, edx
							movd      mm0, eax
							por       mm0, mm5
						}

						MMXSpriteDrawLow(0)
						MMXSpriteIncrementLine()
					}
					return;

				case 1: // xxxxxxxx xxxxxxx1
					MMXSpriteSetColorMask()
					__asm  mov edi, bm

					while (zy--)
					{
						fspr += l_y_skip[l] * dy;

						__asm{
							mov       esi, fspr
							movd      mm0, [esi]
							psrlq     mm0, 28
							por       mm0, mm5
						}

						MMXSpriteDrawLow(0)
						MMXSpriteIncrementLine()
					}
					return;
			}
		}
	}
}

#else

static void NeoMVSDrawGfx(struct osd_bitmap *bitmap,const struct GfxElement *gfx,
		unsigned int code,unsigned int color,int flipx,int flipy,int sx,int sy,
		int zx,int zy,const struct rectangle *clip)
{
	int oy, ey, dy, l = 0;
	UINT32 *fspr;
	UINT16 *bm;

	fspr = (UINT32 *)memory_region(REGION_GFX2);

	/* Mish/AJP - Most clipping is done in main loop */
	oy = sy;
	ey = sy + zy -1; 	/* Clip for size of zoomed object */

	if (sy < clip->min_y) sy = clip->min_y;
	if (ey >= clip->max_y) ey = clip->max_y;

	if (!flipy) /* normal */
	{
		dy = 8;
		fspr+=code*32 + (sy-oy)*2;
	}
	else /* Y flip */
	{
		dy = -8;
		fspr+=(code+1)*32 - 2 - (sy-oy)*2;
	}

	zy = ey - sy + 1;
	bm = &bitmap->line[sy][sx];

	MMXSpriteSetColorMask()

	if (!flipx)	/* normal */
	{
		switch (zx)
		{
			case 16: // 11111111 11111111
					 // 01234567 89ABCDEF
				MMXSpriteStart()
				MMXSpriteSetMask()
X16:			MMXSpriteLoopStart()
				MMXSpriteFullDraw([esi], 0, 8)
				MMXSpriteFullDraw([esi + 4], 16, 24)
				MMXSpriteLoopEnd(X16)
				return;

			case 15: // 1111111x 11111111
				MMXSpriteStart()
				MMXSpriteSetMask()
X15:			MMXSpriteLoopStart()
				__asm {
					mov   eax, [esi]
					mov   edx, [esi + 4]
					mov   ecx, edx
					and   eax, 00fffffffh
					and   ecx, 00000000fh
					shl   ecx, 28
					or    eax, ecx
					shr   edx, 4
				}
				MMXSpriteFullDraw(eax, 0, 8)
				MMXSpriteFullDraw(edx, 16, 24)
				MMXSpriteLoopEnd(X15)
				return;

			case 14: // 1111111x 1111111x
				MMXSpriteStart()
				MMXSpriteSetMask()
X14:			MMXSpriteLoopStart()
				__asm {
					mov   eax, [esi]
					mov   edx, [esi + 4]
					mov   ecx, edx
					and   eax, 00fffffffh
					and   ecx, 00000000fh
					shl   ecx, 28
					or    eax, ecx
					and   edx, 00ffffff0h
					shr   edx, 4
				}
				MMXSpriteFullDraw(eax, 0, 8)
				MMXSpriteFullDraw(edx, 16, 24)
				MMXSpriteLoopEnd(X14)
				return;

			case 13: // 11x1111x 1111x111
				MMXSpriteStart()
				MMXSpriteSetMask()
X13:			MMXSpriteLoopStart()
				__asm {
					mov   eax, [esi]
					mov   edx, [esi + 4]
					mov   ecx, eax
					and   eax, 00ffff000h
					shr   eax, 4
					and   ecx, 0000000ffh
					or    eax, ecx
					mov   ecx, edx
					and   ecx, 0000000ffh
					shl   ecx, 24
					or    eax, ecx
					mov   ecx, edx
					and   ecx, 00000ff00h
					shr   ecx, 8
					and   edx, 0fff00000h
					shr   edx, 12
					or    edx, ecx
				}
				MMXSpriteFullDraw(eax, 0, 8)
				MMXSpriteFullDraw(edx, 16, 24)
				MMXSpriteLoopEnd(X13)
				return;

			case 12: // 111x111x 111x111x
				MMXSpriteStart()
				MMXSpriteSetMask()
X12:			MMXSpriteLoopStart()
				__asm {
					mov   eax, [esi]
					mov   edx, [esi + 4]
					mov   ecx, eax
					and   eax, 00fff0000h
					shr   eax, 4
					and   ecx, 000000fffh
					or    eax, ecx
					mov   ecx, edx
					and   ecx, 0000000ffh
					shl   ecx, 24
					or    eax, ecx
					mov   ecx, edx
					and   ecx, 00fff0000h
					shr   ecx, 12
					and   edx, 000000f00h
					shr   edx, 8
					or    edx, ecx
				}
				MMXSpriteFullDraw(eax, 0, 8)
				MMXSpriteHalfDraw(edx, 16)
				MMXSpriteLoopEnd(X12)
				return;

			case 11: // x11x11x1 11x11x11
				MMXSpriteStart()
				MMXSpriteSetMask()
X11:			MMXSpriteLoopStart()
				__asm {
					mov   eax, [esi]
					mov   edx, [esi + 4]
					mov   ecx, eax
					and   ecx, 000000ff0h
					shr   ecx, 4
					mov   ebx, eax
					and   ebx, 000ff0000h
					shr   ebx, 8
					or    ecx, ebx
					and   eax, 0f0000000h
					shr   eax, 12
					or    eax, ecx
					mov   ecx, edx
					and   ecx, 0000000ffh
					shl   ecx, 20
					or    eax, ecx
					mov   ecx, edx
					and   ecx, 00000f000h
					shl   ecx, 16
					or    eax, ecx
					mov   ecx, edx
					and   ecx, 0000f0000h
					shr   ecx, 16
					and   edx, 0ff000000h
					shr   edx, 20
					or    edx, ecx
				}
				MMXSpriteFullDraw(eax, 0, 8)
				MMXSpriteHalfDraw(edx, 16)
				MMXSpriteLoopEnd(X11)
				return;

			case 10: // 1x1x1x11 1x1x1x11
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X10:			MMXSpriteLoopStart()
				__asm {
					mov       eax, [esi]
					mov       ecx, eax
					and       eax, 00f0f0f0fh
					and       ecx, 0f0000000h
					shr       ecx, 28
					mov       edx, [esi + 4]
					mov       ebx, edx
					and       edx, 0000f0f0fh
					shl       edx, 8
					or        edx, ecx
					mov       ecx, ebx
					and       ecx, 00f000000h
					shr       ecx, 24
					and       ebx, 0f0000000h
					shr       ebx, 12
					or        ecx, ebx
				}
				MMXSpriteUnpackByte(eax, mm0)
				MMXSpriteUnpackByte(edx, mm1)
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				__asm {
					movd      mm0, ecx
					por       mm0, mm5
				}
				MMXSpriteDrawLow(16)
				MMXSpriteLoopEnd(X10)
				return;

			case 9: // 1x1x1x1x 1x1x1x11
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X9:				MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi]
					and       eax, 00f0f0f0fh
					mov       edx, [esi + 4]
					mov       ecx, edx
					and       edx, 00f0f0f0fh
					and       ecx, 0f0000000h
					shr       ecx, 28
				}
				MMXSpriteUnpackByte(eax, mm0)
				MMXSpriteUnpackByte(edx, mm1)
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				__asm {
					movd      mm0, ecx
					por       mm0, mm5
				}
				MMXSpriteDrawLow(16)
				MMXSpriteLoopEnd(X9)
				return;

			case 8: // 1x1x1x1x 1x1x1x1x
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X8:				MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi]
					and       eax, 00f0f0f0fh
					mov       edx, [esi + 4]
					and       edx, 00f0f0f0fh
				}
				MMXSpriteUnpackByte(eax, mm0)
				MMXSpriteUnpackByte(edx, mm1)
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				MMXSpriteLoopEnd(X8)
				return;

			case 7: // x1x1x1xx 1x1x1x1x
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X7:				MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi]
					and       eax, 000f0f0f0h
					shr       eax, 4
					mov       edx, [esi + 4]
					mov       ecx, edx
					and       ecx, 00000000fh
					shl       ecx, 24
					or        eax, ecx
					and       edx, 00f0f0f00h
					shr       edx, 8
				}
				MMXSpriteUnpackByte(eax, mm0)
				MMXSpriteUnpackByte(edx, mm1)
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				MMXSpriteLoopEnd(X7)
				return;

			case 6: // x1xx1xx1 1xx1xx1x
				MMXSpriteStart()
X6:				MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi]
					mov       edx, eax
					mov       ecx, eax
					and       eax, 0000000f0h
					shr       eax, 4
					and       edx, 0000f0000h
					or        eax, edx
					and       ecx, 0f0000000h
					shr       ecx, 28
					mov       edx, [esi + 4]
					mov       ebx, edx
					and       ebx, 00000000fh
					shl       ebx, 16
					or        ecx, ebx
					mov       ebx, edx
					and       ebx, 00000f000h
					shr       ebx, 12
					and       edx, 00f000000h
					shr       edx, 8
					or        edx, ebx
					movd      mm0, eax
					movd      mm1, ecx
					punpckldq mm0, mm1
					psllq     mm1, 32
					por       mm0, mm1
					por       mm0, mm5
					movd      mm1, edx
					por       mm1, mm5
				}
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				MMXSpriteLoopEnd(X6)
				return;

			case 5: // xx1xx1xx 1xx1xx1x
				MMXSpriteStart()
X5:				MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi]
					mov       edx, eax
					and       eax, 000000f00h
					shr       eax, 8
					and       edx, 000f00000h
					shr       edx, 4
					or        eax, edx
					mov       edx, [esi + 4]
					mov       ecx, edx
					mov       ebx, edx
					and       ecx, 00000000fh
					and       ebx, 00000f000h
					shl       ebx, 4
					or        ecx, ebx
					and       edx, 00f000000h
					shr       edx, 24
					movd      mm0, eax
					movd      mm1, ecx
					punpckldq mm0, mm1
					psllq     mm1, 32
					por       mm0, mm1
					por       mm0, mm5
					movd      mm1, edx
					por       mm1, mm5
				}
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				MMXSpriteLoopEnd(X5)
				return;

			case 4: // 1xxx1xxx 1xxx1xxx
				MMXSpriteStart()
X4:				MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi]
					mov       edx, [esi + 4]
					and       eax, 0000f000fh
					and       edx, 0000f000fh
					movd      mm0, eax
					movd      mm1, edx
					psllq     mm1, 32
					por       mm0, mm1
					por       mm0, mm5
				}
				MMXSpriteDrawLow(0)
				MMXSpriteLoopEnd(X4)
				return;

			case 3: // xxx1xxxx 1xxxx1xx
				MMXSpriteStart()
X3:				MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi]
					and       eax, 00000f000h
					shr       eax, 12
					mov       edx, [esi + 4]
					mov       ecx, edx
					and       ecx, 00000000fh
					shl       ecx, 16
					or        eax, ecx
					and       edx, 000f00000h
					shr       edx, 20
					movd      mm0, eax
					movd      mm1, edx
					punpckldq mm0, mm1
					por       mm0, mm5
				}
				MMXSpriteDrawLow(0)
				MMXSpriteLoopEnd(X3)
				return;

			case 2: // 1xxxxxxx 1xxxxxxx
				MMXSpriteStart()
X2:				MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi]
					and       eax, 00000000fh
					mov       edx, [esi + 4]
					and       edx, 00000000fh
					shl       edx, 16
					or        eax, edx
					movd      mm0, eax
					por       mm0, mm5
				}
				MMXSpriteDrawLow(0)
				MMXSpriteLoopEnd(X2)
				return;

			case 1: // 1xxxxxxx xxxxxxxx
				MMXSpriteStart()
X1:				MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi]
					and       eax, 00000000fh
					movd      mm0, eax
					por       mm0, mm5
				}
				MMXSpriteDrawLow(0)
				MMXSpriteLoopEnd(X1)
				return;
		}
	}
	else	/* X flip */
	{
		switch (zx)
		{
			case 16: // 11111111 11111111
					 // FEDCBA98 76543210
				MMXSpriteStart()
				MMXSpriteSetMask()
X16FLIP:		MMXSpriteLoopStart()
				MMXSpriteByteSwap([esi + 4], eax)
				MMXSpriteByteSwap([esi], edx)
				MMXSpriteFullDrawFlipX(eax, 0, 8)
				MMXSpriteFullDrawFlipX(edx, 16, 24)
				MMXSpriteLoopEnd(X16FLIP)
				return;

			case 15: // 11111111 x1111111
				MMXSpriteStart()
				MMXSpriteSetMask()
X15FLIP:		MMXSpriteLoopStart()
				MMXSpriteNibbleSwap([esi + 4], eax)
				MMXSpriteNibbleSwap([esi], edx)
				__asm {
					 shr       edx, 4
				}
				MMXSpriteFullDraw(eax, 0, 8)
				MMXSpriteFullDraw(edx, 16, 24)
				MMXSpriteLoopEnd(X15FLIP)
				return;

			case 14: // x1111111 x1111111
				MMXSpriteStart()
				MMXSpriteSetMask()
X14FLIP:		MMXSpriteLoopStart()
				MMXSpriteNibbleSwap([esi + 4], eax)
				MMXSpriteNibbleSwap([esi], edx)
				__asm {
					mov       ecx, edx
					and       edx, 00ffffff0h
					shr       edx, 4
					shr       eax, 4
					and       ecx, 0000000f0h
					shl       ecx, 24
					or        eax, ecx
				}
				MMXSpriteFullDraw(eax, 0, 8)
				MMXSpriteFullDraw(edx, 16, 24)
				MMXSpriteLoopEnd(X14FLIP)
				return;

			case 13: // 111x1111 x1111x11
				MMXSpriteStart()
				MMXSpriteSetMask()
X13FLIP:		MMXSpriteLoopStart()
				MMXSpriteNibbleSwap([esi + 4], eax)
				MMXSpriteNibbleSwap([esi], edx)
				__asm {
					mov       ecx, edx
					mov       ebx, edx
					and       ecx, 0ff000000h
					shr       ecx, 12
					and       edx, 0000fff00h
					shr       edx, 8
					or        edx, ecx
					and       ebx, 0000000f0h
					shl       ebx, 24
					mov       ecx, eax
					and       ecx, 0ffff0000h
					shr       ecx, 4
					and       eax, 000000fffh
					or        eax, ecx
					or        eax, ebx
				}
				MMXSpriteFullDraw(eax, 0, 8)
				MMXSpriteFullDraw(edx, 16, 24)
				MMXSpriteLoopEnd(X13FLIP)
				return;

			case 12: // x111x111 x111x111
				MMXSpriteStart()
				MMXSpriteSetMask()
X12FLIP:		MMXSpriteLoopStart()
				MMXSpriteNibbleSwap([esi + 4], eax)
				MMXSpriteNibbleSwap([esi], edx)
				__asm {
					mov       ecx, edx
					mov       ebx, edx
					and       ecx, 0fff00000h
					shr       ecx, 16
					and       edx, 00000f000h
					shr       edx, 12
					or        edx, ecx
					and       ebx, 000000ff0h
					shl       ebx, 20
					mov       ecx, eax
					and       ecx, 0fff00000h
					shr       ecx, 8
					and       eax, 00000fff0h
					shr       eax, 4
					or        eax, ecx
					or        eax, ebx
				}
				MMXSpriteFullDraw(eax, 0, 8)
				MMXSpriteHalfDraw(edx, 16)
				MMXSpriteLoopEnd(X12FLIP)
				return;

			case 11: // 11x11x11 1x11x11x
				MMXSpriteStart()
				MMXSpriteSetMask()
X11FLIP:		MMXSpriteLoopStart()
				MMXSpriteNibbleSwap([esi + 4], eax)
				MMXSpriteNibbleSwap([esi], edx)
				__asm {
					mov       ecx, edx
					mov       ebx, edx
					and       ecx, 00ff00000h
					shr       ecx, 16
					and       edx, 00000f000h
					shr       edx, 12
					or        edx, ecx
					mov       ecx, ebx
					and       ecx, 000000f00h
					shl       ecx, 20
					and       ecx, 00000000fh
					shl       ebx, 24
					or        ebx, ecx
					mov       ecx, eax
					and       ecx, 0ff000000h
					shr       ecx, 8
					or        ebx, ecx
					mov       ecx, eax
					and       ecx, 0000ff000h
					shr       ecx, 4
					or        ebx, ecx
					and       eax, 0000000ffh
					or        eax, ebx
				}

				MMXSpriteFullDraw(eax, 0, 8)
				MMXSpriteHalfDraw(edx, 16)
				MMXSpriteLoopEnd(X11FLIP)
				return;

			case 10: // 11x1x1x1 11x1x1x1
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X10FLIP:		MMXSpriteLoopStart()
				MMXSpriteByteSwap([esi + 4], eax)
				MMXSpriteByteSwap([esi], edx)
				__asm {
					mov       ecx, eax
					mov       ebx, eax
					and       ecx, 0000f0f0fh
					shl       ecx, 8
					and       eax, 0000000f0h
					shr       eax, 4
					or        eax, ecx
					and       ebx, 00f000000h
					shr       ebx, 24
					mov       ecx, edx
					and       edx, 0000000f0h
					shl       edx, 4
					or        edx, ebx
					mov       ebx, ecx
					and       ebx, 000000f0fh
					shl       ebx, 16
					or        edx, ebx
					mov       ebx, ecx
					and       ebx, 0000f0000h
					shr       ebx, 16
					and       ecx, 00f000000h
					shr       ecx, 8
					or        ecx, ebx
				}
				MMXSpriteUnpackByte(eax, mm0)
				MMXSpriteUnpackByte(edx, mm1)
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				__asm {
					movd      mm0, ecx
					por       mm0, mm5
				}
				MMXSpriteDrawLow(16)
				MMXSpriteLoopEnd(X10FLIP)
				return;

			case 9:  // 11x1x1x1 x1x1x1x1
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X9FLIP:			MMXSpriteLoopStart()
				MMXSpriteByteSwap([esi + 4], eax)
				MMXSpriteByteSwap([esi], edx)
				__asm {
					mov       ecx, eax
					mov       ebx, eax
					and       ecx, 0000f0f0fh
					shl       ecx, 8
					and       eax, 0000000f0h
					shr       eax, 4
					or        eax, ecx
					and       ebx, 00f000000h
					shr       ebx, 24
					mov       ecx, edx
					and       edx, 0000f0f0fh
					shl       edx, 8
					or        edx, ebx
					and       ecx, 00f000000h
					shr       ecx, 24
				}
				MMXSpriteUnpackByte(eax, mm0)
				MMXSpriteUnpackByte(edx, mm1)
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				__asm {
					movd      mm0, ecx
					por       mm0, mm5
				}
				MMXSpriteDrawLow(16)
				MMXSpriteLoopEnd(X9FLIP)
				return;

			case 8:  // x1x1x1x1 x1x1x1x1
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X8FLIP:			MMXSpriteLoopStart()
				MMXSpriteByteSwap([esi + 4], eax)
				MMXSpriteByteSwap([esi], edx)
				__asm {
					and       eax, 0f0f0f0fh
					and       edx, 0f0f0f0fh
				}
				MMXSpriteUnpackByte(eax, mm0)
				MMXSpriteUnpackByte(edx, mm1)
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				MMXSpriteLoopEnd(X8FLIP)
				return;

			case 7: // x1x1x1x1 xx1x1x1x
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X7FLIP:			MMXSpriteLoopStart()
				MMXSpriteByteSwap([esi + 4], eax)
				MMXSpriteByteSwap([esi], edx)
				__asm {
					and       eax, 00f0f0f0fh
					and       edx, 0f0f0f000h
					shr       edx, 12
				}
				MMXSpriteUnpackByte(eax, mm0)
				MMXSpriteUnpackByte(edx, mm1)
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				MMXSpriteLoopEnd(X7FLIP)
				return;

			case 6: // x1xx1xx1 1xx1xx1x
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X6FLIP:			MMXSpriteLoopStart()
				__asm {
					mov       eax, [esi + 4]
					mov       edx, eax
					mov       ecx, eax
					and       eax, 00f000000h
					shr       eax, 24
					and       edx, 00000f000h
					shr       edx, 4
					or        eax, edx
					and       ecx, 00000000fh
					shl       ecx, 16
					or        eax, ecx
					mov       edx, [esi]
					mov       ebx, edx
					and       ebx, 0f0000000h
					shr       ebx, 4
					or        eax, ebx
					mov       ecx, edx
					and       edx, 0000f0000h
					shr       edx, 16
					and       ecx, 0000000f0h
					shl       ecx, 12
					or        edx, ecx
					movd      mm1, edx
					por       mm1, mm5
				}
				MMXSpriteUnpackByte(eax, mm0)
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				MMXSpriteLoopEnd(X6FLIP)
				return;

			case 5: // x1xx1xx1 xx1xx1xx
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X5FLIP:			MMXSpriteLoopStart()
				__asm {
					mov       eax, [esi + 4]
					mov       edx, eax
					mov       ecx, eax
					and       eax, 00f000000h
					shr       eax, 24
					and       edx, 00000f000h
					shr       edx, 4
					or        eax, edx
					and       ecx, 00000000fh
					shl       ecx, 16
					or        eax, ecx
					mov       edx, [esi]
					mov       ebx, edx
					and       ebx, 000f00000h
					shl       ebx, 4
					or        eax, ebx
					and       edx, 000000f00h
					shr       edx, 8
					movd      mm1, edx
					por       mm1, mm5
				}
				MMXSpriteUnpackByte(eax, mm0)
				MMXSpriteDrawLow(0)
				MMXSpriteDrawHigh(8)
				MMXSpriteLoopEnd(X5FLIP)
				return;

			case 4: // xxx1xxx1 xxx1xxx1
				MMXSpriteStart()
X4FLIP:			MMXSpriteLoopStart()
				MMXSpriteByteSwap([esi + 4], eax)
				MMXSpriteByteSwap([esi], edx)
				__asm{
					and       eax, 00f000f00h
					shr       eax, 8
					and       edx, 00f000f00h
					shr       edx, 8
					movd      mm0, eax
					movd      mm1, edx
					psllq     mm1, 32
					por       mm0, mm1
					por       mm0, mm5
				}
				MMXSpriteDrawLow(0)
				MMXSpriteLoopEnd(X4FLIP)
				return;

			case 3: // xx1xxxx1 xxxx1xxx
				MMXSpriteStart()
				MMXSpriteSetZeroMask()
X3FLIP:			MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi + 4]
					mov       edx, eax
					and       eax, 000f00000h
					shr       eax, 20
					and       edx, 00000000fh
					shl       eax, 8
					or        eax, edx
					mov       edx, [esi]
					and       edx, 00000f000h
					shl       edx, 4
					or        eax, edx
					movd      mm0, eax
					punpcklbw mm0, mm6
					por       mm0, mm5
				}
				MMXSpriteDrawLow(0)
				MMXSpriteLoopEnd(X3FLIP)
				return;

			case 2: // xxxxxxx1 xxxxxxx1
				MMXSpriteStart()
X2FLIP:			MMXSpriteLoopStart()
				__asm{
					mov       eax, [esi + 4]
					and       eax, 0f0000000h
					shr       eax, 28
					mov       edx, [esi]
					and       edx, 0f0000000h
					shr       edx, 12
					or        eax, edx
					movd      mm0, eax
					por       mm0, mm5
				}
				MMXSpriteDrawLow(0)
				MMXSpriteLoopEnd(X2FLIP)
				return;

			case 1: // xxxxxxxx xxxxxxx1
				MMXSpriteStart()
X1FLIP:			MMXSpriteLoopStart()
				__asm{
					movd      mm0, [esi]
					psrlq     mm0, 28
					por       mm0, mm5
				}
				MMXSpriteDrawLow(0)
				MMXSpriteLoopEnd(X1FLIP)
				return;
		}
	}
}
#endif

#else
/******************************************************************************/

const static UINT32 MASK28 = 0xf0000000;
const static UINT32 MASK24 = 0x0f000000;
const static UINT32 MASK20 = 0x00f00000;
const static UINT32 MASK16 = 0x000f0000;
const static UINT32 MASK12 = 0x0000f000;
const static UINT32 MASK08 = 0x00000f00;
const static UINT32 MASK04 = 0x000000f0;
const static UINT32 MASK00 = 0x0000000f;

#define PAL(x)   (color|(x))

static void NeoMVSDrawGfx(struct osd_bitmap *bitmap,const struct GfxElement *gfx,
		unsigned int code,unsigned int color,int flipx,int flipy,int sx,int sy,
		int zx,int zy,const struct rectangle *clip)
{
	int oy, ey, dy;
	UINT32 *fspr;

	/* Mish/AJP - Most clipping is done in main loop */
	oy = sy;
	ey = sy + zy -1; 	/* Clip for size of zoomed object */

	if (sy < clip->min_y) sy = clip->min_y;
	if (ey >= clip->max_y) ey = clip->max_y;

	fspr = (UINT32 *)memory_region(REGION_GFX2);

	if (!flipy) /* normal */
	{
		dy = 2;
		fspr+=code*32 + (sy-oy)*2;
	}
	else /* Y flip */
	{
		dy = -2;
		fspr+=(code+1)*32 - 2 - (sy-oy)*2;
	}

	zy = ey - sy + 1;
	color <<= 4;

	{
		UINT32 col, col4;
		int l = 0;
		int dm = bitmap->modulo;
		UINT16 *bm = &bitmap->line[sy][sx];

		if (!flipx)	/* normal */
		{
			switch (zx)
			{
				case 16:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 1) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+ 3) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+ 4) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 5) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 6) = PAL(col>>24);
						col = (col4 & MASK28); if (col) *(bm+ 7) = PAL(col>>28);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 8) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 9) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+10) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+11) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+12) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+13) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+14) = PAL(col>>24);
						col = (col4 & MASK28); if (col) *(bm+15) = PAL(col>>28);
						l++;
						bm += dm;
					}
					return;

				case 15:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 1) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+ 3) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+ 4) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 5) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 6) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 7) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 8) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 9) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+10) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+11) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+12) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+13) = PAL(col>>24);
						col = (col4 & MASK28); if (col) *(bm+14) = PAL(col>>28);
						l++;
						bm += dm;
					}
					return;

				case 14:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 1) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+ 3) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+ 4) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 5) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 6) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 7) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 8) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 9) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+10) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+11) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+12) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+13) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 13:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 1) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 3) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 4) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 5) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 6) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 7) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 8) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+ 9) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+10) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+11) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+12) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 12:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 1) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 3) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 4) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 5) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 6) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 7) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 8) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 9) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+10) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+11) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 11:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 1) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 3) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 4) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 5) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 6) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 7) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 8) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 9) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+10) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 10:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 1) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 3) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 4) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 5) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 6) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 7) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 8) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 9) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 9:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 1) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 2) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 3) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 4) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 5) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 6) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 7) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 8) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 8:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 1) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 2) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 3) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 4) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 5) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 6) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 7) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 7:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 1) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 2) = PAL(col>>16);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 3) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 4) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 5) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 6) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 6:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 1) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 2) = PAL(col>>16);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 3) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 4) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 5) = PAL(col>>16);
						l++;
						bm += dm;
					}
					return;

				case 5:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK16); if (col) *(bm+ 1) = PAL(col>>16);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 2) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 3) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 4) = PAL(col>>16);
						l++;
						bm += dm;
					}
					return;

				case 4:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col = (col4 & MASK16); if (col) *(bm+ 1) = PAL(col>>16);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 2) = PAL(col    );
						col = (col4 & MASK16); if (col) *(bm+ 3) = PAL(col>>16);
						l++;
						bm += dm;
					}
					return;

				case 3:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 1) = PAL(col    );
						col = (col4 & MASK16); if (col) *(bm+ 2) = PAL(col>>16);
						l++;
						bm += dm;
					}
					return;

				case 2:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 1) = PAL(col    );
						l++;
						bm += dm;
					}
					return;

				case 1:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						l++;
						bm += dm;
					}
					return;
			}
		}
		else
		{
			switch (zx)
			{
				case 16:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+15) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+14) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+13) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+12) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+11) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+10) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 9) = PAL(col>>24);
						col = (col4 & MASK28); if (col) *(bm+ 8) = PAL(col>>28);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 7) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 6) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 5) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+ 4) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+ 3) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 2) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 1) = PAL(col>>24);
						col = (col4 & MASK28); if (col) *(bm+ 0) = PAL(col>>28);
						l++;
						bm += dm;
					}
					return;

				case 15:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+14) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+13) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+12) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+11) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+10) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 9) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 8) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 7) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 6) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 5) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+ 4) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+ 3) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 2) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 1) = PAL(col>>24);
						col = (col4 & MASK28); if (col) *(bm+ 0) = PAL(col>>28);
						l++;
						bm += dm;
					}
					return;

				case 14:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+13) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+12) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+11) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+10) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+ 9) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 8) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 7) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 6) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 5) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 4) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+ 3) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+ 2) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 1) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 0) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 13:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+12) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+11) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+10) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 9) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 8) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 7) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 6) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 5) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 4) = PAL(col>> 8);
						col = (col4 & MASK12); if (col) *(bm+ 3) = PAL(col>>12);
						col = (col4 & MASK16); if (col) *(bm+ 2) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 1) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 0) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 12:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+11) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+10) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 9) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 8) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 7) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 6) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 5) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 4) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 3) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 2) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 1) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 0) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 11:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+10) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 9) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 8) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 7) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 6) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 5) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 4) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 3) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 2) = PAL(col>>16);
						col = (col4 & MASK20); if (col) *(bm+ 1) = PAL(col>>20);
						col = (col4 & MASK24); if (col) *(bm+ 0) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 10:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 9) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 8) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 7) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 6) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 5) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 4) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 3) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 1) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 0) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 9:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 8) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 7) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 6) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 5) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 4) = PAL(col    );
						col = (col4 & MASK04); if (col) *(bm+ 3) = PAL(col>> 4);
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 1) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 0) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 8:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 7) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 6) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 5) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 4) = PAL(col>>24);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 3) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 1) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 0) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 7:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 6) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 5) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 4) = PAL(col>>16);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 3) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 2) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 1) = PAL(col>>16);
						col = (col4 & MASK24); if (col) *(bm+ 0) = PAL(col>>24);
						l++;
						bm += dm;
					}
					return;

				case 6:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 5) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 4) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 3) = PAL(col>>16);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 2) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 1) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 0) = PAL(col>>16);
						l++;
						bm += dm;
					}
					return;

				case 5:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 4) = PAL(col    );
						col = (col4 & MASK16); if (col) *(bm+ 3) = PAL(col>>16);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 2) = PAL(col    );
						col = (col4 & MASK08); if (col) *(bm+ 1) = PAL(col>> 8);
						col = (col4 & MASK16); if (col) *(bm+ 0) = PAL(col>>16);
						l++;
						bm += dm;
					}
					return;

				case 4:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 3) = PAL(col    );
						col = (col4 & MASK16); if (col) *(bm+ 2) = PAL(col>>16);
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 1) = PAL(col    );
						col = (col4 & MASK16); if (col) *(bm+ 0) = PAL(col>>16);
						l++;
						bm += dm;
					}
					return;

				case 3:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 2) = PAL(col    );
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 1) = PAL(col    );
						col = (col4 & MASK16); if (col) *(bm+ 0) = PAL(col>>16);
						l++;
						bm += dm;
					}
					return;

				case 2:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *fspr;
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 1) = PAL(col    );
						l++;
						bm += dm;
					}
					return;

				case 1:
					while (zy--)
					{
						fspr += l_y_skip[l] * dy;
						col4 = *(fspr + 1);
						col = (col4 & MASK00); if (col) *(bm+ 0) = PAL(col    );
						l++;
						bm += dm;
					}
					return;
			}
		}
	}
}
#endif

void NeoMVSDrawGfx16(struct osd_bitmap *bitmap,const struct GfxElement *gfx,
		unsigned int code,unsigned int color,int flipx,int flipy,int sx,int sy,
		int zx,int zy,const struct rectangle *clip)
{
	/* Safety feature */
	code = code % no_of_tiles;

	if (!gfx->pen_usage[code])	/* decode tile if it hasn't been yet */
		decodetile(code);

	/* Check for total transparency, no need to draw */
	if (gfx->pen_usage[code] & ~1)
	{
		NeoMVSDrawGfx(bitmap,
			gfx,
			code,
			color,
			flipx, flipy,
			sx,sy,zx,zy,
			clip
		);
	}
}

/******************************************************************************/

#ifdef MMX_FIXCHAR
static void NeoDrawGfxChar16(struct osd_bitmap *bitmap, const struct GfxElement *gfx,
	unsigned int code, unsigned int color, int sx, int sy)
{
	UINT16 *dd    = &bitmap->line[sy][sx];
	UINT8  *sd    = gfx->gfxdata + code * gfx->char_modulo;
	int    opaque = gfx->pen_usage[code];

	__asm {
		movd      mm7, color
		punpcklwd mm7, mm7
		punpckldq mm7, mm7
		pxor      mm6, mm6

		mov       eax, sd
		mov       edx, dd
		mov       ecx, 8

		and       opaque, 1
		jz        L2

L1:
		movd      mm0, [eax]
		movq      mm1, [edx]
		punpcklbw mm0, mm6
		por       mm0, mm7
		movq      mm2, mm0
		pcmpeqw   mm2, mm7
		pand      mm1, mm2
		pandn     mm2, mm0
		por       mm2, mm1
		movq      [edx], mm2

		movd      mm0, [eax + 4]
		movq      mm1, [edx + 8]
		punpcklbw mm0, mm6
		por       mm0, mm7
		movq      mm2, mm0
		pcmpeqw   mm2, mm7
		pand      mm1, mm2
		pandn     mm2, mm0
		por       mm2, mm1
		movq      [edx + 8], mm2

		lea       eax, [eax + 8]
		lea       edx, [edx + (320 + 16) * 2]
		dec       ecx
		jnz       L1
		jmp       short L3
L2:
		movd      mm0, [eax]
		punpcklbw mm0, mm6
		por       mm0, mm7
		movq      [edx], mm0

		movd      mm1, [eax + 4]
		punpcklbw mm1, mm6
		por       mm1, mm7
		movq      [edx + 8], mm1

		lea       eax, [eax + 8]
		lea       edx, [edx + (320 + 16) * 2]
		dec       ecx
		jnz       L2
L3:
	}
}
#else
#define read_dword(address) *(int *)address

#ifdef LSB_FIRST
#define BL0 0
#define BL1 1
#define BL2 2
#define BL3 3
#else
#define BL0 3
#define BL1 2
#define BL2 1
#define BL3 0
#endif

const static UINT32 MASKX24 = 0xff000000;
const static UINT32 MASKX16 = 0x00ff0000;
const static UINT32 MASKX08 = 0x0000ff00;
const static UINT32 MASKX00 = 0x000000ff;

INLINE void NeoDrawGfxChar16(struct osd_bitmap *dest, const struct GfxElement *gfx,
		unsigned int code, unsigned int color, int sx, int sy)
{
	int y = 8;
	int dm = dest->modulo;
	UINT16 *dd = &dest->line[sy][sx];
	UINT8 *sd = gfx->gfxdata + code * gfx->char_modulo;

	if (!(gfx->pen_usage[code] & 1))
	{
		/* character is totally opaque, can disable transparency */

		while (y--)
		{
			*(dd+0) = PAL(*sd++);
			*(dd+1) = PAL(*sd++);
			*(dd+2) = PAL(*sd++);
			*(dd+3) = PAL(*sd++);
			*(dd+4) = PAL(*sd++);
			*(dd+5) = PAL(*sd++);
			*(dd+6) = PAL(*sd++);
			*(dd+7) = PAL(*sd++);
			dd += dm;
		}
	}
	else
	{
		int *sd4 = (int *)sd;
		int col, col4;

		dm -= 4;

		while (y--)
		{
			if ((col4 = read_dword(sd4)))
			{
				col = (col4 & MASKX00); if (col) *(dd+BL0) = PAL(col    );
				col = (col4 & MASKX08); if (col) *(dd+BL1) = PAL(col>> 8);
				col = (col4 & MASKX16); if (col) *(dd+BL2) = PAL(col>>16);
				col = (col4 & MASKX24); if (col) *(dd+BL3) = PAL(col>>24);
			}
			dd += 4;
			sd4++;
			if ((col4 = read_dword(sd4)))
			{
				col = (col4 & MASKX00); if (col) *(dd+BL0) = PAL(col    );
				col = (col4 & MASKX08); if (col) *(dd+BL1) = PAL(col>> 8);
				col = (col4 & MASKX16); if (col) *(dd+BL2) = PAL(col>>16);
				col = (col4 & MASKX24); if (col) *(dd+BL3) = PAL(col>>24);
			}
			dd += dm;
			sd4++;
		}
	}
}
#endif


/******************************************************************************/

static void neogeo_draw_sfix(struct osd_bitmap *bitmap)
{
	int x, y;
	int code, color;
	unsigned int *pen_usage;
	struct GfxElement *gfx;
	extern int front_layer;

	if (front_layer)	return;
	
	gfx = Machine->gfx[fix_bank];
	pen_usage = gfx->pen_usage;

	/* Character foreground */
	for (y = 1; y < 30; y++)
	{
		for (x = 1; x < 40; x++)
		{
			code = neogeo_vidram16[(0xe000 >> 1) + y + 32 * x];
			color = (code >> 8) & 0x00f0;

			code &= 0x0fff;

			if (fix_bank && !neogeo_machine_country)
			{
				if (sdodgeb_fix)
				{
					switch (code)
					{
					case 0x016e:
					case 0x0166:
					case 0x0173:
						code -= 0xc0;
						break;

					case 0x026e:
					case 0x0273:
					case 0x0266:
						code -= 0x180;
						break;

					case 0x010d:
						code = 0x008d;
						break;
					case 0x020d:
						code = 0x00cd;
						break;

					default:
						break;
					}
				}

				if (code >= 0x100 && code <= 0x2ff)
					code += 0x200;
			}

			if (!(pen_usage[code] & ~1)) continue;

			NeoDrawGfxChar16(
				bitmap,
				gfx,
				code,
				color,
				x << 3,
				y << 3
			);
		}
	}
}

/******************************************************************************/

#ifdef MMX_CLEAR
/* fill a bitmap using the specified pen with clip */
static void fillbitmap_yclip(struct osd_bitmap *dest, UINT32 pen, const struct rectangle *clip)
{
	int sy, ey;
	UINT16 *pDst;

	sy = clip->min_y;
	ey = clip->max_y - sy + 1;
	pDst = &dest->line[sy][8];

	__asm {
		mov       edi, pDst
		mov       edx, ey
		mov       eax, (16 + 16) * 2
		movd      mm0, pen
		punpckldq mm0, mm0
L1:
		mov       ecx, (320 - 16) / 4
L2:
		movq      [edi], mm0
		dec       ecx
		lea       edi, [edi + 8]
		jnz       L2
		dec       edx
		lea       edi, [edi + eax]
		jnz       L1
	}
}
#else
/* fill a bitmap using the specified pen with clip */
static void fillbitmap_yclip(struct osd_bitmap *dest, UINT32 pen, const struct rectangle *clip)
{
	int sy, ey;
	UINT16 *pDst;

	sy = clip->min_y;
	ey = clip->max_y - sy + 1;
	pDst = &dest->line[sy][8];

#if 0
	__asm {
		mov       edi, pDst
		mov       edx, ey
		mov       ebx, (16 + 16) * 2
L1:
		mov       ecx, (320 - 16) / 2
L2:
		mov       [edi], eax
		dec       ecx
		lea       edi, [edi + 4]
		jnz       L2
		dec       edx
		lea       edi, [edi + ebx]
		jnz       L1
	}
#else
	__asm {
		mov       edi, pDst
		mov       eax, pen
		mov       edx, ey
		mov       ebx, (16 + 16) * 2
L1:
		mov       ecx, (320 - 16) / 2
		rep       stosd
		dec       edx
		lea       edi, [edi + ebx]
		jnz       L1
	}
#endif
}
#endif


/******************************************************************************/

static void screenrefresh(struct osd_bitmap *bitmap, const struct rectangle *clip)
{
	int sx = 0, sy = 0, oy = 0, my = 0, zx = 0, zy = 0;
	int offs, count, y;
	int tileno, tileatr, t1, t3;
	char fullmode = 0;
	struct GfxElement *gfx = Machine->gfx[2]; /* Save constant struct dereference */
	extern int skip_object[0x300 >> 1];

	if (clip->max_y - clip->min_y > 8 ||	/* kludge to speed up raster effects */
		clip->min_y == Machine->visible_area.min_y)
    {
		/* Palette swap occured after last frame but before this one */
		if (palette_swap_pending) swap_palettes();
	}

	fillbitmap_yclip(bitmap, 0x0fff0fff, clip);

	object_number = 0;

	for (count = 0; count < 0x300 >> 1; count++)
	{
		t3 = neogeo_vidram16[(0x10000 >> 1) + count];
		t1 = neogeo_vidram16[(0x10400 >> 1) + count];

		/* If this bit is set this new column is placed next to last one */
		if (t1 & 0x40)
		{
			/* No point doing anything if tile strip is 0 */
			if (my == 0) continue;

			sx += zx;
			/* Get new zoom for this column */
			zx = (t3 >> 8) & 0x0f;
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
				zx = 0x0f;
			zx++;
			sy = oy;
		}
		else /* nope it is a new block */
		{
			/* Number of tiles in this strip */
			my = t1 & 0x3f;
			if (my == 0)
				continue; /* No point doing anything if tile strip is 0 */
			else if (my == 0x20)
				fullmode = 1;
			else if (my >= 0x21)
				fullmode = 2;	/* most games use 0x21, but  Alpha Mission II uses 0x3f */
			else
				fullmode = 0;

			/* Sprite scaling */
			zy = t3 & 0xff;
			/* Get new X zoom for this column */
			zx = (t3 >> 8) & 0x0f;
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
			{
				zx = 0x0f;
				zy = 0xff;
			}
			if (zy == 0)
			{
				/* No point doing anything if tile Y zoom is 0 */
				my = 0;
				continue;
			}
			zx++;

			sy = 0x200 - (t1 >> 7);

			if (clip->max_y - clip->min_y > 8 ||	/* kludge to improve the ssideki games */
				clip->min_y == Machine->visible_area.min_y)
			{
				if (sy > 0x110) sy -= 0x200;
				if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
				{
					while (sy < 0) sy += 2 * (zy + 1);
				}
			}
			oy = sy;

			if (zy < 0xff && my < 0x10)
				my = mmy[zy][my];
			else if (my > 0x20)
				my = 0x20;

			sx = (neogeo_vidram16[(0x10800 >> 1) + count] >> 7);
		}

		if (sx >= 320)
			sx -= 0x200;

		if (sx + zx <= 8)
			continue;

		offs = count << 6;

		if (!(t1 & 0x40))	object_number++;

		if (skip_object[object_number])	continue;

		/* my holds the number of tiles in each vertical multisprite block */
		for (y = 0; y < my ; y++)
		{
			if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
			{
				if (sy >= 248) sy -= 2 * (zy + 1);
			}
			else if (fullmode == 1)
			{
				if (y == 0x10) sy -= 2 * (zy + 1);
			}
			else if (sy > 0x110)
			{
				sy -= 0x200;	/* NS990105 mslug2 fix */
			}

			if (sy + yskip[zy][y] > clip->min_y && sy <= clip->max_y)
			{
				tileno  = neogeo_vidram16[offs++];
				tileatr = neogeo_vidram16[offs++];

				if (high_tile && tileatr & 0x10)   tileno += 0x10000;
				if (vhigh_tile && tileatr & 0x20)  tileno += 0x20000;
				if (vvhigh_tile && tileatr & 0x40) tileno += 0x40000;

				if (tileatr & 0x08)
				{
					tileno = (tileno & ~7) + (neogeo_frame_counter & 7);
				}
				else if (tileatr & 0x04)
				{
					tileno = (tileno & ~3) + (neogeo_frame_counter & 3);
				}

				l_y_skip = dda_y_skip[zy][y];

				/* Safety feature */
				tileno = tileno % no_of_tiles;

				if (!gfx->pen_usage[tileno])	/* decode tile if it hasn't been yet */
					decodetile(tileno);

				/* Check for total transparency, no need to draw */
				if (gfx->pen_usage[tileno] & ~1)
				{
					NeoMVSDrawGfx(bitmap,
						gfx,
						tileno,
						tileatr >> 8,
						tileatr & 0x01,tileatr & 0x02,
						sx,sy,zx,yskip[zy][y],
						clip
					);
				}
			}
			else offs += 2;

			sy += yskip[zy][y];
		}  /* for y */
	}  /* for count */
}

/******************************************************************************/

void neogeo_vh_screenrefresh(struct osd_bitmap *bitmap,int full_refresh)
{
#ifdef NEOGEO_MMX
//		__asm emms    ; FPU -> MMX (for K6)
#endif
	screenrefresh(bitmap, &Machine->visible_area);
	neogeo_draw_sfix(bitmap);
#ifdef NEOGEO_MMX
	__asm emms    ; MMX -> FPU
#endif
}

/******************************************************************************/

static int next_update_first_line;

void neogeo_vh_raster_partial_refresh(struct osd_bitmap *bitmap,int current_line)
{
	struct rectangle clip;

	if (current_line < next_update_first_line)
		next_update_first_line = 0;

	clip.min_x = Machine->visible_area.min_x;
	clip.max_x = Machine->visible_area.max_x;
	clip.min_y = next_update_first_line;
	clip.max_y = current_line;
	if (clip.min_y < Machine->visible_area.min_y)
		clip.min_y = Machine->visible_area.min_y;
	if (clip.max_y > Machine->visible_area.max_y)
		clip.max_y = Machine->visible_area.max_y;

	if (clip.max_y >= clip.min_y)
	{
#ifdef NEOGEO_MMX
//		__asm emms    ; FPU -> MMX (for K6)
#endif
//		logerror("refresh %d-%d\n",clip.min_y,clip.max_y);
		screenrefresh(bitmap, &clip);
#ifdef NEOGEO_MMX
		__asm emms    ; MMX -> FPU
#endif
	}

	next_update_first_line = current_line + 1;
}

void neogeo_vh_raster_screenrefresh(struct osd_bitmap *bitmap,int full_refresh)
{
	extern int paused;

	if (!paused)
	{
#ifdef NEOGEO_MMX
//		__asm emms    ; FPU -> MMX (for K6)
#endif
		neogeo_draw_sfix(bitmap);
#ifdef NEOGEO_MMX
		__asm emms    ; MMX -> FPU
#endif
	}
	else
//		neogeo_restore_screen();
		neogeo_vh_screenrefresh(bitmap, full_refresh);
}

/******************************************************************************/

static void raster_screenrefresh1(struct osd_bitmap *bitmap, const struct rectangle *fullclip, const struct rectangle *partclip)
{
	int sx = 0, sy = 0, oy = 0, my = 0, zx = 0, zy = 0;
	int offs, count, y;
	int tileno, tileatr, t1, t3;
	int fullmode = 0, firsttime = 1;
	struct GfxElement *gfx = Machine->gfx[2]; /* Save constant struct dereference */
	const struct rectangle *clip;

	/* Draw sprites */
	for (count = 0; count < 0x300 >> 1; count++)
	{
		t3 = neogeo_vidram16[(0x10000 >> 1) + count];
		t1 = neogeo_vidram16[(0x10400 >> 1) + count];

		/* If this bit is set this new column is placed next to last one */
		if (t1 & 0x40)
		{
			/* No point doing anything if tile strip is 0 */
			if (my == 0) continue;

			sx += zx;
			/* Get new zoom for this column */
			zx = (t3 >> 8) & 0x0f;
//			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
//				zx = 0x0f;
			zx++;
			sy = oy;
		}
		else
		{
			/* nope it is a new block */

			/* Number of tiles in this strip */
			my = t1 & 0x3f;
			if (my == 0)
				continue; /* No point doing anything if tile strip is 0 */
			else if (my == 0x20)
				fullmode = 1;
			else if (my >= 0x21)
				fullmode = 2;	/* most games use 0x21, but  Alpha Mission II uses 0x3f */
			else
				fullmode = 0;

			if (fullmode && firsttime)
			{
				struct rectangle fillclip;

				fillclip.min_y = raster_start_line;
				fillclip.max_y = 239; //Machine->visible_area.max_y;
				fillbitmap_yclip(bitmap, 0xffffffff, &fillclip);

				firsttime = 0;
			}

			/* Sprite scaling */
			zy = t3 & 0xff;
			/* Get new X zoom for this column */
			zx = (t3 >> 8) & 0x0f;
#if 0
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
			{
				zx = 0x0f;
				zy = 0xff;
			}
#endif
			if (zy == 0)
			{
				/* No point doing anything if tile Y size is 0 */
				my = 0;
				continue;
			}
			zx++;

			sy = 0x200 - (t1 >> 7);
			if (partclip->max_y - partclip->min_y > 8 ||	/* kludge to improve the ssideki games */
				partclip->min_y == Machine->visible_area.min_y)
			{
				if (sy > 0x110) sy -= 0x200;
				if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
				{
					while (sy < 0) sy += 2 * (zy + 1);
				}
			}
			oy = sy;

			if (zy < 0xff && my < 0x10)
				my = mmy[zy][my];
			else if (my > 0x20)
				my = 0x20;

			sx = (neogeo_vidram16[(0x10800 >> 1) + count]) >> 7;
		}

		if (sx >= 312) sx -= 0x200;

		if (sx + zx <= 8) continue;

		if (fullmode)
			clip = partclip;
		else
			clip = fullclip;

		offs = count << 6;

		/* my holds the number of tiles in each vertical multisprite block */
		for (y = 0; y < my ; y++)
		{
			if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
			{
				if (sy >= 248) sy -= 2 * (zy + 1);
			}
			else if (fullmode == 1)
			{
				if (y == 0x10) sy -= 2 * (zy + 1);
			}
#if 0
			else if (sy > 0x110)
			{
				sy -= 0x200;	/* NS990105 mslug2 fix */
			}
#endif

			if (sy + yskip[zy][y] > clip->min_y && sy <= clip->max_y)
			{
				tileno  = neogeo_vidram16[offs++];
				tileatr = neogeo_vidram16[offs++];

				if (high_tile && tileatr & 0x10)   tileno += 0x10000;
				if (vhigh_tile && tileatr & 0x20)  tileno += 0x20000;
				if (vvhigh_tile && tileatr & 0x40) tileno += 0x40000;

				if (tileatr & 0x08)
				{
					tileno = (tileno & ~7) + (neogeo_frame_counter & 7);
				}
				else if (tileatr & 0x04)
				{
					tileno = (tileno & ~3) + (neogeo_frame_counter & 3);
				}

				l_y_skip = dda_y_skip[zy][y];

				/* Safety feature */
				tileno = tileno % no_of_tiles;

				if (!gfx->pen_usage[tileno])	/* decode tile if it hasn't been yet */
					decodetile(tileno);

				/* Check for total transparency, no need to draw */
				if (gfx->pen_usage[tileno] & ~1)
				{
					NeoMVSDrawGfx(bitmap,
						gfx,
						tileno,
						tileatr >> 8,
						tileatr & 0x01,tileatr & 0x02,
						sx,sy,zx,yskip[zy][y],
						clip
					);
				}
			}
			else offs += 2;

			sy += yskip[zy][y];
		}  /* for y */
	}  /* for count */
}

static void raster_screenrefresh2(struct osd_bitmap *bitmap, const struct rectangle *clip)
{
	int sx=0, sy=0, oy=0, my=0, zx=0, zy=0;
	int offs, count = 0, y;
	int tileno, tileatr, t1, t3;
	int fullmode = 0;
	struct GfxElement *gfx = Machine->gfx[2]; /* Save constant struct dereference */

	/* Skip drawing */
#if 0
seek_next:
	sx = sy = oy = my = zx = zy = fullmode = 0;
#endif

	for (; count < 0x300 >> 1; count++)
	{
		t1 = neogeo_vidram16[(0x10400 >> 1) + count];
		if (/*!(t1 & 0x40) &&*/ (t1 & 0x3f) >= 0x20)
			break;
	}

	/* Draw sprites */
	for (; count < 0x300 >> 1; count++)
	{
		t3 = neogeo_vidram16[(0x10000 >> 1) + count];
		t1 = neogeo_vidram16[(0x10400 >> 1) + count];

		/* If this bit is set this new column is placed next to last one */
		if (t1 & 0x40)
		{
			/* No point doing anything if tile strip is 0 */
			if (my == 0) continue;

			sx += zx;
			/* Get new zoom for this column */
			zx = (t3 >> 8) & 0x0f;
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
				zx = 0x0f;
			zx++;
			sy = oy;
		}
		else
		{
			/* nope it is a new block */

			/* Number of tiles in this strip */
			my = t1 & 0x3f;
			if (my == 0x20)
				fullmode = 1;
			else if (my >= 0x21)
				fullmode = 2;	/* most games use 0x21, but  Alpha Mission II uses 0x3f */
			else
#if 0
			{
				my = 0;
				count++;
				goto seek_next;
			}
#else
			return;
#endif

			/* Sprite scaling */
			zy = t3 & 0xff;
			/* Get new X zoom for this column */
			zx = (t3 >> 8) & 0x0f;
#if 0
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
			{
				zx = 0x0f;
				zy = 0xff;
			}
#endif
			if (zy == 0)
#if 0
			{
				my = 0;
				count++;
				goto seek_next;
			}
#else
			return;
#endif
			zx++;

			sy = 0x200 - (t1 >> 7);
#if 0
			if (clip->max_y - clip->min_y > 8 ||	/* kludge to improve the ssideki games */
				clip->min_y == Machine->visible_area.min_y)
			{
				if (sy > 0x110) sy -= 0x200;
				if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
				{
					while (sy < 0) sy += 2 * (zy + 1);
				}
			}
#endif
			oy = sy;

			if (zy < 0xff && my < 0x10)
				my = mmy[zy][my];
			else if (my > 0x20)
				my = 0x20;

			sx = neogeo_vidram16[(0x10800 >> 1) + count] >> 7;
		}

		if (sx >= 312) sx -= 0x200;

		if (sx + zx <= 8) continue;

//		if (!fullmode) continue;

		offs = count << 6;

		/* my holds the number of tiles in each vertical multisprite block */
		for (y = 0; y < my; y++)
		{
			if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
			{
				if (sy >= 248) sy -= 2 * (zy + 1);
			}
			else if (fullmode == 1)
			{
				if (y == 0x10) sy -= 2 * (zy + 1);
			}
#if 0
			else if (sy > 0x110)
			{
				sy -= 0x200;	/* NS990105 mslug2 fix */
			}
#endif

			if (sy + yskip[zy][y] > clip->min_y && sy <= clip->max_y)
			{
				tileno	= neogeo_vidram16[offs++];
				tileatr = neogeo_vidram16[offs++];

				if (high_tile && tileatr & 0x10)   tileno += 0x10000;
				if (vhigh_tile && tileatr & 0x20)  tileno += 0x20000;
				if (vvhigh_tile && tileatr & 0x40) tileno += 0x40000;

				if (tileatr & 0x08)
				{
					tileno = (tileno & ~7) + (neogeo_frame_counter & 7);
				}
				else if (tileatr & 0x04)
				{
					tileno = (tileno & ~3) + (neogeo_frame_counter & 3);
				}

				l_y_skip = dda_y_skip[zy][y];

				/* Safety feature */
				tileno = tileno % no_of_tiles;

				if (!gfx->pen_usage[tileno])	/* decode tile if it hasn't been yet */
					decodetile(tileno);

				/* Check for total transparency, no need to draw */
				if (gfx->pen_usage[tileno] & ~1)
				{
					NeoMVSDrawGfx(bitmap,
						gfx,
						tileno,
						tileatr >> 8,
						tileatr & 0x01,tileatr & 0x02,
						sx,sy,zx,yskip[zy][y],
						clip
					);
				}
			}
			else offs += 2;

			sy += yskip[zy][y];
		}  /* for y */
	}  /* for count */
}

void neogeo_vh_raster2_partial_refresh(struct osd_bitmap *bitmap,int current_line)
{
	struct rectangle clip;

	clip.min_x = 16; //Machine->visible_area.min_x;
	clip.max_x = 239; //Machine->visible_area.max_x;

	if (current_line < next_update_first_line)
	{
		next_update_first_line = 0;
		raster_start_line = current_line + 1;

		if (palette_swap_pending) swap_palettes();

		if (raster_start_line <= 239)
		{
			fillbitmap(bitmap, 0x0fff0fff);

			clip.min_y = raster_start_line;
			clip.max_y = 239; //Machine->visible_area.max_y;
			fillbitmap_yclip(sub_bitmap, 0x0fff0fff, &clip);
		}
	}

	clip.min_y = next_update_first_line;
	clip.max_y = current_line;
	if (clip.min_y < Machine->visible_area.min_y)
		clip.min_y = Machine->visible_area.min_y;
	if (clip.max_y > Machine->visible_area.max_y)
		clip.max_y = Machine->visible_area.max_y;

#ifdef NEOGEO_MMX
//	__asm emms    ; FPU -> MMX (for K6)
#endif

	if (!next_update_first_line)
	{
		if (raster_start_line <= 239)
			raster_screenrefresh1(bitmap, &Machine->visible_area, &clip);
		else
			screenrefresh(bitmap, &Machine->visible_area);
	}
	else if (clip.max_y >= clip.min_y)
	{
//		logerror("refresh %d-%d\n",clip.min_y,clip.max_y);
		raster_screenrefresh2(sub_bitmap, &clip);
	}

#ifdef NEOGEO_MMX
	__asm emms    ; MMX -> FPU
#endif

	next_update_first_line = current_line + 1;
}

#ifdef MMX_MARGEBMP
void neogeo_vh_raster2_screenrefresh(struct osd_bitmap *bitmap,int full_refresh)
{
	extern int paused;

	if (!paused)
	{
//		__asm emms    ; FPU -> MMX (for K6)

		if (raster_start_line <= 239)
		{
			int    nHeight = Machine->visible_area.max_y - raster_start_line + 1;
			UINT16 *pSrc   = &sub_bitmap->line[raster_start_line][8];
			UINT16 *pDst   = &bitmap->line[raster_start_line][8];

			__asm {
				mov       edx, nHeight
				mov       eax, pSrc
				mov       edi, pDst
				pcmpeqw   mm3, mm3
L1:
				mov       ecx, (320 - 16) / 4
L2:
				movd      mm0, [eax]
				movd      mm1, [eax + 4]
				punpckldq mm0, mm1
				movd      mm1, [edi]
				movd      mm2, [edi + 4]
				punpckldq mm1, mm2
				movq      mm2, mm1
				pcmpeqw   mm2, mm3
				pand      mm0, mm2
				pandn     mm2, mm1
				por       mm0, mm2
				movq      [edi], mm0
				lea       eax, [eax + 8]
				lea       edi, [edi + 8]
				dec       ecx
				jnz       L2

				lea       eax, [eax + 64]
				lea       edi, [edi + 64]
				dec       edx
				jnz       L1
			}
		}

		neogeo_draw_sfix(bitmap);

		__asm emms    ; MMX -> FPU

		raster_start_line = 240;
	}
	else
		neogeo_vh_screenrefresh(bitmap, full_refresh);
//		neogeo_restore_screen();
}
#else
void neogeo_vh_raster2_screenrefresh(struct osd_bitmap *bitmap,int full_refresh)
{
	extern int paused;

	if (!paused)
	{
		if (raster_start_line <= 239)
		{
			int x, y;

			for (y = raster_start_line; y <= Machine->visible_area.max_y; y++)
			{
				for (x = Machine->visible_area.min_x; x <= Machine->visible_area.max_x; x++)
				{
					if (bitmap->line[y][x] == 0xffff)
						bitmap->line[y][x] = sub_bitmap->line[y][x];
				}
			}
		}

#ifdef NEOGEO_MMX
//	__asm emms    ; FPU -> MMX (for K6)
#endif

		neogeo_draw_sfix(bitmap);

#ifdef NEOGEO_MMX
		__asm emms    ; MMX -> FPU
#endif

		raster_start_line = 240;
	}
	else
		neogeo_vh_screenrefresh(bitmap, full_refresh);
//		neogeo_restore_screen();
}
#endif


void neogeo_save_screen(void)
{
	if (Machine->gamedrv->flags & GAME_RASTER_DRIVER)
	{
		int x, y;
		neogeo_vh_screenrefresh(Machine->scrbitmap, 1);

		for (y = 16; y < 240; y++)
			for (x = 8; x < 312; x++)
				sub_bitmap->line[y][x] = Machine->scrbitmap->line[y][x];
	}
}

void neogeo_restore_screen(void)
{
	if (Machine->gamedrv->flags & GAME_RASTER_DRIVER)
	{
		int x, y;

		for (y = 16; y < 240; y++)
			for (x = 8; x < 312; x++)
				Machine->scrbitmap->line[y][x] = sub_bitmap->line[y][x];
	}
}


/******************************************************************************/

#define NMAME_HEADER     "NMSTA10\0"
#define NMAME_HEADER_V09 "NMSTA09\0"

#define CPU_TYPE(index)  (Machine->drv->cpu[index].cpu_type & ~CPU_FLAGS_MASK)

/*-----------------------------------------------------------------------------*/
/* current context size                                                        */
/*-----------------------------------------------------------------------------*/

#define STATE_HEADER_SIZE        26
#define MC68K_CONTEXT_SIZE       29
#define Z80_CONTEXT_SIZE         20
#define YM2610_CONTEXT_SIZE      2545
#define NEOGEO_CONTEXT_SIZE      78
#define NEOGEO_CONTEXT_SIZE_V09  270
#define MISC_CONTEXT_SIZE    0x10000 + 0x10c00 + PALETTE_SIZE * 2 \
                             + sizeof(int) + 0x800

#define CONTEXT_SIZE         MC68K_CONTEXT_SIZE * sizeof(int) \
                             + Z80_CONTEXT_SIZE * sizeof(int) \
                             + YM2610_CONTEXT_SIZE \
                             + NEOGEO_CONTEXT_SIZE \
                             + MISC_CONTEXT_SIZE

#define CONTEXT_SIZE_V09     MC68K_CONTEXT_SIZE * sizeof(int) \
                             + Z80_CONTEXT_SIZE * sizeof(int) \
                             + YM2610_CONTEXT_SIZE \
                             + NEOGEO_CONTEXT_SIZE_V09 \
                             + MISC_CONTEXT_SIZE

/*-----------------------------------------------------------------------------*/

void get_date_and_time(void);
void neogeo_get_context(char *buf);
void neogeo_set_context(char *buf, int version);
void ym2610_get_context(char *buf);
void ym2610_set_context(char *buf, int version);
char *GetStateDir(void);

int saveslot = 0;
char date_str[9];
char time_str[9];
char info_str[20];

int save_contexts(int filetype, const int slot)
{
	void *f;
	char fname[128];

	if (slot == TEMPORARY_FILE)
		sprintf(fname, "temp.sta");
	else
		sprintf(fname, "%s.%03d", Machine->gamedrv->name, slot);

	if ((f = osd_fopen(0, fname, filetype, 1)) != NULL)
	{
		int  context_68k[MC68K_CONTEXT_SIZE];
		int  context_z80[Z80_CONTEXT_SIZE];
		char context_ym2610[YM2610_CONTEXT_SIZE];
		char context_neogeo[NEOGEO_CONTEXT_SIZE];
		int i, bank = 0;
		data16_t *mem16 = (data16_t *)memory_region(REGION_CPU2);


		get_date_and_time();

		for (i = 1; i <= MC68K_CONTEXT_SIZE; i++)
			context_68k[i - 1] = cpuintf[CPU_TYPE(0)].get_reg(i);

		for (i = 1; i <= Z80_CONTEXT_SIZE; i++)
			context_z80[i - 1] = cpuintf[CPU_Z80].get_reg(i);

		if (neogeo_paletteram16 == neogeo_palettebank[1])
			bank = 1;

		ym2610_get_context(context_ym2610);
		neogeo_get_context(context_neogeo);

		osd_fwrite(f, NMAME_HEADER, 8);
		osd_fwrite(f, date_str, 9);
		osd_fwrite(f, time_str, 9);
		osd_fwrite(f, context_68k,           MC68K_CONTEXT_SIZE * sizeof(int));
		osd_fwrite(f, context_z80,           Z80_CONTEXT_SIZE * sizeof(int));
		osd_fwrite(f, context_ym2610,        YM2610_CONTEXT_SIZE);
		osd_fwrite(f, context_neogeo,        NEOGEO_CONTEXT_SIZE);
		osd_fwrite(f, neogeo_ram16,          0x10000);
		osd_fwrite(f, neogeo_vidram16,       0x10c00);
		osd_fwrite(f, neogeo_palettebank[0], PALETTE_SIZE);
		osd_fwrite(f, neogeo_palettebank[1], PALETTE_SIZE);
		osd_fwrite(f, &bank,                 sizeof(int));
		osd_fwrite(f, &mem16[0xf800 >> 1],   0x800);
		osd_fclose(f);

		logerror("state saved.\n");

		return 1;
	}

	logerror("save state failed.\n");

	return 0;
}


int restore_contexts(int filetype, const int slot)
{
	void *f;
	char fname[128], buf[8];

	if (slot == TEMPORARY_FILE)
		sprintf(fname, "temp.sta");
	else
		sprintf(fname, "%s.%03d", Machine->gamedrv->name, slot);

	if ((f = osd_fopen(0, fname, filetype, 0)) != NULL)
	{
		osd_fread(f, buf, 8);

		if (!strcmp(buf, NMAME_HEADER) || !strcmp(buf, NMAME_HEADER_V09))
		{
			int  context_68k[MC68K_CONTEXT_SIZE];
			int  context_z80[Z80_CONTEXT_SIZE];
			char context_ym2610[YM2610_CONTEXT_SIZE];
			char context_neogeo[NEOGEO_CONTEXT_SIZE_V09];	// ݊邽
			int size_68k, size_z80, size_ym2610, size_neogeo;
			int i, bank = 0, version;
			data16_t *mem16 = (data16_t *)memory_region(REGION_CPU2);
			extern UINT16 combocount[2];

			version = atoi(buf + 5);

			switch (version)
			{
				case 10:
					size_68k    = MC68K_CONTEXT_SIZE * sizeof(int);
					size_z80    = Z80_CONTEXT_SIZE * sizeof(int);
					size_ym2610 = YM2610_CONTEXT_SIZE;
					size_neogeo = NEOGEO_CONTEXT_SIZE;
					break;

				case 9:
					size_68k    = MC68K_CONTEXT_SIZE * sizeof(int);
					size_z80    = Z80_CONTEXT_SIZE * sizeof(int);
					size_ym2610 = YM2610_CONTEXT_SIZE;
					size_neogeo = NEOGEO_CONTEXT_SIZE_V09;
					break;

				default:
					logerror("load state failed.\n");
					osd_fclose(f);
					return 1;
			}

			osd_fseek(f, 18, SEEK_CUR);
			osd_fread(f, context_68k,           size_68k);
			osd_fread(f, context_z80,           size_z80);
			osd_fread(f, context_ym2610,        size_ym2610);
			osd_fread(f, context_neogeo,        size_neogeo);
			osd_fread(f, neogeo_ram16,          0x10000);
			osd_fread(f, neogeo_vidram16,       0x10c00);
			osd_fread(f, neogeo_palettebank[0], PALETTE_SIZE);
			osd_fread(f, neogeo_palettebank[1], PALETTE_SIZE);
			osd_fread(f, &bank,                 sizeof(int));
			osd_fread(f, &mem16[0xf800 >> 1],   0x800);
			osd_fclose(f);

			for (i = 1; i <= MC68K_CONTEXT_SIZE; i++)
				cpuintf[CPU_TYPE(0)].set_reg(i, context_68k[i - 1]);

			for (i = 1; i <= Z80_CONTEXT_SIZE; i++)
				cpuintf[CPU_Z80].set_reg(i, context_z80[i - 1]);

			neogeo_paletteram16 = neogeo_palettebank[bank];
			palette_swap_pending = 1;

			for (i = 0; i < 0x2000 >> 1; i++)
				neogeo_paletteram16_w(i, neogeo_palettebank[bank][i], 0);

			neogeo_set_context(context_neogeo, version);
			ym2610_set_context(context_ym2610, version);

			combocount[0] = combocount[1] = 0;

			logerror("state restored.\n");

			return 1;
		}

		osd_fclose(f);
	}

	logerror("load state failed.\n");

	return 0;
}


int check_state_data(void)
{
	void *f;
	char fname[128], buf[8];
	int retcode = 0;
	long fsize = 0;
	int state_version = 0;

	strcpy(date_str, "--/--/--");
	strcpy(time_str, "--:--:--");

	sprintf(fname,"%s.%03d", Machine->gamedrv->name, saveslot);

	if ((f = osd_fopen(0, fname, OSD_FILETYPE_STATE, 0)) != NULL)
	{
		osd_fseek(f, 0, SEEK_END);
		fsize = osd_ftell(f);
		osd_fseek(f, 0, SEEK_SET);
		osd_fread(f, buf, 8);
		osd_fread(f, date_str, 9);
		osd_fread(f, time_str, 9);
		osd_fclose(f);

		if (strcmp(buf, NMAME_HEADER) && strcmp(buf, NMAME_HEADER_V09)) fsize = 1;
	}

	if (fsize)
	{
		state_version = atoi(buf + 5);
		sprintf(info_str, "NMAME State (Ver.%d.%d)",
			state_version / 10, state_version % 10);

		if (state_version == 10 && fsize == CONTEXT_SIZE + STATE_HEADER_SIZE)
			retcode = 1;
		else if (state_version == 9 && fsize == CONTEXT_SIZE_V09 + STATE_HEADER_SIZE)
			retcode = 1;
		else
			strcpy(info_str, "Unknown Data");
	}
	else
	{
		strcpy(info_str, "No Data");
	}

	return retcode;
}

int count_object_number(void)
{
	return object_number;
}

void alt_backcolor_exp(void)
{
	extern int alt_backcolor;

	if (alt_backcolor)
	{
		palette_change_color(0x0fff, 0xff, 0x00, 0xff); //ALT_BACKGROUND_COLOR
	}
	else
	{
		data16_t newword = neogeo_paletteram16[0x0fff];
#ifndef USE_RGBTABLE
		int r, g, b;

		r = ((newword >> 7) & 0x1e) | ((newword >> 14) & 0x01);
		g = ((newword >> 3) & 0x1e) | ((newword >> 13) & 0x01);
		b = ((newword << 1) & 0x1e) | ((newword >> 12) & 0x01);

		r = (r << 3) | (r >> 2);
		g = (g << 3) | (g >> 2);
		b = (b << 3) | (b >> 2);

		palette_change_color(0x0fff, r, g, b);
#else
		palette_change_color(0x0fff, RGB[newword][0], RGB[newword][1], RGB[newword][2]);
#endif
	}
}
